unit sfGamma;

{Common code for special functions: Gamma function and related}

interface

{$i std.inc}

{$ifdef BIT16}
{$N+}
{$endif}

{$ifdef NOBASM}
  {$undef BASM}
{$endif}

(*************************************************************************

 DESCRIPTION   :  Common code for special functions: Gamma function and related

 REQUIREMENTS  :  TP5.5-7, D2-D7/D9-D10/D12/D17-D18, FPC, VP, WDOSX

 EXTERNAL DATA :  ---

 MEMORY USAGE  :  ---

 DISPLAY MODE  :  ---

 REMARK        :  - The unit can be compiled with TP5 but some functions generate
                    invalid operations due to TP5's brain-damaged usage of the FPU
                  - The unit cannot be compiled with D1 (Error 48: Code segment too large)

 REFERENCES    :  References used in this unit, main index in sfBasic

                  [7] Cephes Mathematical Library, Version 2.8
                      http://www.moshier.net/#Cephes or http://www.netlib.org/cephes/
                 [14] SLATEC Common Mathematical Library, Version 4.1, July 1993
                      (general purpose mathematical and statistical routines written in Fortran 77)
                      http://www.netlib.org/slatec
                 [19] Boost C++ Libraries, Release 1.42.0, 2010.
                      http://www.boost.org/
                 [26] N.M. Temme, A Set of Algorithms for the Incomplete Gamma Functions,
                      Probability in the Engineering and Informational Sciences, 8 (1994), pp.291-307,
                      available as http://oai.cwi.nl/oai/asset/10080/10080A.pdf
                 [27] A.R. Didonato, A.H. Morris, Computation of the Incomplete Gamma Function Ratios
                      and their Inverse. ACM TOMS, Vol 12, No 4, Dec 1986, pp.377-393.
                      Fortran source: ACM TOMS 13 (1987), pp.318-319; available from
                      http://netlib.org/toms/654
                 [32] D.E. Knuth: The Art of computer programming; Volume 1, Fundamental Algorithms,
                      3rd ed., 1997; Volume 2, Seminumerical Algorithms, 3rd ed., 1998;
                      http://www-cs-faculty.stanford.edu/~knuth/taocp.html
                 [36] P. Borwein, An Efficient Algorithm for the Riemann Zeta Function,
                      CMS Conference Proc. 27 (2000), pp. 29-34. Available as
                      http://www.cecm.sfu.ca/personal/pborwein/PAPERS/P155.pdf
                 [37] A.R. DiDonato, A.H. Morris, Algorithm 708: Significant digit computation
                      of the incomplete beta function ratios, ACM TOMS, Vol.18, No.3, 1992,
                      pp.360-373. Fortran source available from http://netlib.org/toms/708

 Version  Date      Author      Modification
 -------  --------  -------     ------------------------------------------
 1.00.00  17.08.10  W.Ehrhardt  Common version number after split
 1.00.01  17.08.10  we          Temme's gstar function
 1.00.02  23.08.10  we          sfc_incgamma, sfc_igammap, sfc_igammaq
 1.00.03  25.08.10  we          sfc_incgamma: cases when a is integer or half integer
 1.00.04  29.08.10  we          sfc_incgamma: dax calculated via Lanczos sum
 1.00.05  01.09.10  we          sfc_incgamma_ex, sfc_igprefix
 1.00.06  04.09.10  we          sfc_incgamma_inv
 1.00.07  05.09.10  we          sfc_incgamma_inv: use eps_d for p+q=1 check
 1.00.08  05.09.10  we          sfc_incgamma_inv: return Inf if p=1 and q=0
 1.00.09  06.09.10  we          sfc_incgamma_inv: Bug fixed code for DM Eq (34)
 1.00.10  08.09.10  we          Improved arg checking and NAN/INF handling
 1.00.11  10.09.10  we          sfc_trigamma
 1.00.12  11.09.10  we          ibeta_series with iteration limit and convergence check
 1.00.13  14.09.10  we          sfc_eta, sfc_etam1
 1.00.14  15.09.10  we          sfc_zetam1

 1.01.00  06.10.10  we          sfc_dfac
 1.01.01  18.10.10  we          Lanczos sum with and without expg scale
 1.01.02  19.10.10  we          sfc_gamma_delta_ratio, sfc_gamma_ratio, sfc_pochhammer
 1.01.03  23.10.10  we          sfc_zetah

 1.02.00  11.11.10  we          sfc_beta uses sfc_pochhammer for integer x+y=0
 1.02.01  11.11.10  we          Fixed missing check in sfc_lnbeta

 1.03.00  07.12.10  we          Fixed sfc_psi(+INF) = +INF
 1.03.01  13.12.10  we          sfc_trigamma with Hurwitz zeta
 1.03.02  14.12.10  we          sfc_tetragamma, sfc_pentagamma
 1.03.03  15.12.10  we          sfc_polygamma
 1.03.04  19.12.10  we          sfc_(ln)gamma return NAN/RTE if x is a non-positive integer
 1.03.05  29.12.10  we          improved sfc_dfac for a few small values to return integer
 1.03.06  30.12.10  we          sfc_binomial

 1.08.00  03.07.11  we          Asymptotic expansion for x>1024 in stirf
 1.08.01  12.07.11  we          polygam_negx: polygamma for x<0 and 1<=n<=10
 1.08.02  03.08.11  we          invgamma renamed to rgamma
 1.08.03  04.08.11  we          small simplification in sfc_gstar
 1.08.04  06.08.11  we          sfc_dfac for odd negative n, fac/dfac return INF if undefined
 1.08.05  08.08.11  we          const ETAEPS, improved etam1pos
 1.08.06  09.08.11  we          improved sfc_zetah

 1.09.00  14.09.11  we          improved lngamma_small
 1.09.01  14.09.11  we          sfc_lngamma1p
 1.09.02  17.09.11  we          lanczos interfaced
 1.09.03  17.09.11  we          sfc_ibetaprefix
 1.09.04  18.09.11  we          improved sfc_ibeta
 1.09.05  06.10.11  we          sfc_zetah for 0 <= s < 1

 1.10.00  24.11.11  we          sfc_lnfac
 1.10.01  15.12.11  we          sfc_igprefix for TP5 without Lanczos
 1.10.02  21.12.11  we          TP5 for sfc_zetah/sfc_zetam1
 1.10.03  22.12.11  we          etam1pos with exp7

 1.11.00  03.02.12  we          Moved zeta and eta functions to sfZeta

 1.12.00  21.03.12  we          sfc_gamma_delta_ratio for negative arguments
 1.12.01  21.03.12  we          improved sfc_pochhammer
 1.12.02  22.03.12  we          sfc_poch1
 1.12.03  09.04.12  we          fix special case if prefix is zero in sfc_incgamma_ex
 1.12.04  11.04.12  we          sfc_igamma (non-normalised upper incomplete gamma)
 1.12.05  18.04.12  we          sfc_lngammas
 1.12.06  22.04.12  we          sfc_gamma for x < -MAXGAMX
 1.12.07  22.04.12  we          sfc_igammal (non-normalised lower incomplete gamma)

 1.13.00  07.06.12  we          fix sfc_igprefix for a<1 and underflow of exp(-x)
 1.13.01  08.06.12  we          improved sfc_rgamma
 1.13.02  09.06.12  we          sfc_taylor
 1.13.03  28.06.12  we          igaml_mser
 1.13.04  30.06.12  we          sfc_git (Tricomi's incomplete gamma)
 1.13.05  01.07.12  we          check loss of accuracy in sfc_igammal

 1.14.00  20.07.12  we          removed special treatment for a=0 in sfc_poch

 1.15.00  15.02.13  we          improved sfc_git
 1.15.01  16.02.13  we          fix sfc_trigamma/tetragamma/pentagamma for very large arguments

 1.16.00  30.03.13  we          improved sfc_binomial

 1.17.00  15.04.13  we          improved sfc_beta
 1.17.01  16.04.13  we          ibeta_series with parameter normalised
 1.17.02  17.04.13  we          sfc_nnbeta
 1.17.03  21.04.13  we          lanczos_gm05 = lanczos_g - 0.5 in implementation
 1.17.04  27.04.13  we          polygam_negx for n=11,12

 1.18.00  11.05.13  we          improved sfc_polygamma for large order: e.g. (3000,200)
 1.18.01  19.05.13  we          Prevent some wrong compiler optimizations for div by 3
 1.18.02  26.05.13  we          sfc_nnbeta for a<=0 or b<=0

***************************************************************************)


(*-------------------------------------------------------------------------
 (C) Copyright 2009-2013 Wolfgang Ehrhardt

 This software is provided 'as-is', without any express or implied warranty.
 In no event will the authors be held liable for any damages arising from
 the use of this software.

 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:

 1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software in
    a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

 2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

 3. This notice may not be removed or altered from any source distribution.
----------------------------------------------------------------------------*)

(*-------------------------------------------------------------------------
  This Pascal code uses material and ideas from open source and public
  domain libraries, see the file '3rdparty.ama' for the licenses.
---------------------------------------------------------------------------*)

const
  MAXGAMX = 1755.455;  {max. argument for sfc_gamma}
  XLGAE   = 8.0;       {threshold for asymptotic expansion of sfc_lngamma}


function sfc_lngcorr(x: extended): extended;
  {-Return lngamma correction term lngamma(x) - ((x-0.5)*ln(x) - x + ln(sqrt(2*Pi)), x>=8}

function sfc_rgamma(x: extended): extended;
  {-Return the reciprocal gamma function rgamma = 1/gamma(x)}

function sfc_binomial(n,k: integer; dbl: boolean): extended;
  {-Return the binomial coefficient 'n choose k'; set dbl=true, if used in SpecFun}

function sfc_fac(n: integer): extended;
  {-Return the factorial n!, n<MAXGAMX-1; INF if n<0}

function sfc_lnfac(n: longint): extended;
  {-Return ln(n!), INF if n<0}

function sfc_dfac(n: integer): extended;
  {-Return the double factorial n!!, n<=3209; INF for even n<0}

function sfc_gamma(x: extended): extended;
  {-Return gamma(x), |x| <= MAXGAMX; NAN/RTE if x is a non-positive integer}

function sfc_gamma_medium(x: extended): extended;
  {-Return gamma(x), |x| <= 13, x negative integer produces div by 0}

function sfc_gaminv_small(x: extended): extended;
  {- 1/gamma(x) for small argument values, |x| < 0.03125}

function sfc_gamma1pm1(x: extended): extended;
  {-Return gamma(1+x)-1 with increased accuracy for x near 0}

function sfc_gamma_delta_ratio(x,d: extended): extended;
  {-Return gamma(x)/gamma(x+d), accurate even for |d| << |x|}

function sfc_gamma_ratio(x,y: extended): extended;
  {-Return gamma(x)/gamma(y)}

function sfc_pochhammer(a,x: extended): extended;
  {-Return the Pochhammer symbol (a)_x = gamma(a+x)/gamma(a).}

function sfc_poch1(a,x: extended): extended;
  {-Return (pochhammer(a,x)-1)/x, psi(a) if x=0; accurate even for small |x|}

function sfc_lngamma(x: extended): extended;
  {-Return ln(|gamma(x)|), |x| <= MAXLGM, NAN/RTE if x is a non-positive integer}

function sfc_lngammas(x: extended; var s: integer): extended;
  {-Return ln(|gamma(x)|), |x| <= MAXLGM, s=-1,1 is the sign of gamma}

function sfc_lngamma1p(x: extended): extended;
  {-Return ln(|gamma(1+x)|) with increased accuracy for x near 0}

function sfc_signgamma(x: extended): extended;
  {-Return sign(gamma(x)), useless for 0 or negative integer}

function sfc_gstar(x: extended): extended;
  {-Return Temme's gstar(x) = gamma(x)/(sqrt(2*Pi)*x^(x-0.5)*exp(-x)), x>0.}
  { For large x the asymptotic expansion is gstar(x) = 1 + 1/12x + O(1/x^2)}

function sfc_igprefix(a,x: extended): extended;
  {-return x^a*exp(-x)/gamma(a) using Lanczos sum}

procedure sfc_incgamma_ex(a,x: extended; var p,q,dax: extended; calcdax: boolean);
  {-Return the normalised incomplete gamma functions P and Q, a>=0, x>=0}
  { P(a,x) = integral(exp(-t)*t^(a-1), t=0..x  )/gamma(a)}
  { Q(a,x) = integral(exp(-t)*t^(a-1), t=x..Inf)/gamma(a)}
  { If calcdax=true, dax = x^a*exp(-x)/gamma(a) is returned. This is extra}
  { effort only in the special cases, otherwise dax is calculated anyway.}

procedure sfc_incgamma(a,x: extended; var p,q: extended);
  {-Return the normalised incomplete gamma functions P and Q, a>=0, x>=0}
  { P(a,x) = integral(exp(-t)*t^(a-1), t=0..x  )/gamma(a)}
  { Q(a,x) = integral(exp(-t)*t^(a-1), t=x..Inf)/gamma(a)}

function sfc_igammap(a,x: extended): extended;
  {-Return the normalised lower incomplete gamma function P(a,x), a>=0, x>=0}
  { P(a,x) = integral(exp(-t)*t^(a-1), t=0..x)/gamma(a)}

function sfc_igammaq(a,x: extended): extended;
  {-Return the normalised upper incomplete gamma function Q(a,x), a>=0, x>=0}
  { Q(a,x) = integral(exp(-t)*t^(a-1), t=x..Inf)/gamma(a)}

function sfc_igamma(a,x: extended): extended;
  {-Return the non-normalised upper incomplete gamma function}
  { GAMMA(a,x) = integral(exp(-t)*t^(a-1), t=x..Inf), x>=0}

function sfc_igammal(a,x: extended): extended;
  {-Return the non-normalised lower incomplete gamma function}
  { gamma(a,x) = integral(exp(-t)*t^(a-1), t=0..x); x>=0, a<>0,-1,-2,..}

function sfc_git(a,x: extended): extended;
  {-Return Tricomi's entire incomplete gamma function, x>=0,}
  { gammastar(a,x) = igammal(a,x)/gamma(a)/x^a = P(a,x)/x^a }

procedure sfc_incgamma_inv(a,p,q: extended; var x: extended; var ierr: integer);
  {-Return the inverse normalised incomplete gamma function, i.e. calculate}
  { x with P(a,x)=p and Q(a,x)=q. Input parameter a>0, p>=0, q>0 and p+q=1.}
  { ierr is >= 0 for success, < 0 for input errors or iterations failures. }

function sfc_igamma_inv(a,p,q: extended): extended;
  {-Return the inverse normalised incomplete gamma function, i.e. calculate}
  { x with P(a,x)=p and Q(a,x)=q. Input parameter a>0, p>=0, q>0 and p+q=1.}

function sfc_igammap_inv(a,p: extended): extended;
  {-Inverse incomplete gamma: return x with P(a,x)=p, a>=0, 0<=p<1}

function sfc_igammaq_inv(a,q: extended): extended;
  {-Inverse complemented incomplete gamma: return x with Q(a,x)=q, a>=0, 0<q<=1}

function sfc_psi(x: extended): extended;
  {-Return the psi (digamma) function of x, INF if x is a non-positive integer}

function sfc_trigamma(x: extended): extended;
  {-Return the trigamma function psi'(x), INF if x is a negative integer}

function sfc_tetragamma(x: extended): extended;
  {-Return the tetragamma function psi''(x), NAN/RTE if x is a negative integer}

function sfc_pentagamma(x: extended): extended;
  {-Return the pentagamma function psi'''(x), INF if x is a negative integer}

function sfc_polygamma(n: integer; x: extended): extended;
  {-Return the polygamma function: n'th derivative of psi; n>=0, x>0 for n>12.}
  { Note: Accuracy may be reduced for n>=MAXGAMX due to ln/exp operations.}

function sfc_beta(x,y: extended): extended;
  {-Return the function beta(x,y)=gamma(x)*gamma(y)/gamma(x+y)}

function sfc_lnbeta(x,y: extended): extended;
  {-Return the logarithm of |beta(x,y)|=|gamma(x)*gamma(y)/gamma(x+y)|}

function sfc_ibeta(a, b, x: extended): extended;
  {-Return the normalised incomplete beta function, a>0, b>0, 0 <= x <= 1}
  { sfc_ibeta = integral(t^(a-1)*(1-t)^(b-1) / beta(a,b), t=0..x)}

function sfc_ibetaprefix(a,b,x,y: extended): extended;
  {-Return (x^a)(y^b)/Beta(a,b)}

function sfc_nnbeta(a, b, x: extended): extended;
  {-Return the non-normalised incomplete beta function B_x(a,b)}
  { for 0<=x<=1, B_x = integral(t^(a-1)*(1-t)^(b-1), t=0..x).  }

function sfc_taylor(x: extended; n: integer): extended;
  {-Return the Taylor coefficient x^n/n!, n>=0}

{#Z+}
procedure igam_qfraction(a,x,dax: extended; var p,q: extended);
  {-Temme/Gautschi continued fraction for Q(a,x)}
{#Z-}


implementation

uses
  AMath,
  sfBasic,  {Basic common code}
  sfZeta,   {Zeta functions and polylogarithms}
  sfExpInt, {Exponential integrals and related}
  sfHyperG, {Hypergeometric functions}
  sfErf;    {Error function and related}

const
{ lanczos_g = 13.144565 + 5.45696821063756942749e-17;}
  lgm05_hex : THexExtW = ($0000,$B257,$2363,$CA50,$4002);  {12.6445650000000000545696821}
var
  lanczos_gm05: extended absolute lgm05_hex; {lanczos_g - 0.5}

{---------------------------------------------------------------------------}
function lanczos(x: extended; egscale: boolean): extended;
  {-Return the Lanczos sum for x, exp(g) scaled if egscale=true}
const
  lsh:  array[0..12] of THexExtW = (
          ($6F92,$D992,$7F81,$A01D,$402C),  {+4.4012138428004608955E+13}
          ($CD54,$4004,$20CB,$974E,$402C),  {+4.1590453358593200516E+13}
          ($F967,$596F,$659A,$8311,$402B),  {+1.8013842787117996778E+13}
          ($71D8,$A66C,$D4A5,$899F,$4029),  {+4.7287362634753888969E+12}
          ($96F3,$2C67,$5538,$C317,$4026),  {+8.3791008362840464705E+11}
          ($817F,$4B70,$3C24,$C4AA,$4023),  {+1.0558370727342993449E+11}
          ($125B,$9FAE,$C3E8,$908F,$4020),  {+9.7013636184949994935E+9}
          ($B883,$315C,$CD76,$9C24,$401C),  {+6.5491439754820526409E+8}
          ($29BA,$97D5,$7978,$F5F5,$4017),  {+3.2238322942133565306E+7}
          ($EA59,$87B0,$11C1,$89C2,$4013),  {+1.1285142194970914380E+6}
          ($0534,$EF56,$966A,$D053,$400D),  {+2.6665793784598589447E+4}
          ($8931,$781E,$A7EE,$BEF0,$4007),  {+3.8188012486329268705E+2}
          ($2CB3,$B138,$98FF,$A06C,$4000)); {+2.5066282746310005025}
  lseh: array[0..12] of THexExtW = (
          ($EBF0,$180B,$E131,$A434,$4019),  {+8.6091529534185372176E+7}
          ($A3D3,$B6ED,$E125,$9B2B,$4019),  {+8.1354505178580112428E+7}
          ($78B7,$D786,$C498,$866A,$4018),  {+3.5236626388154619108E+7}
          ($F7DC,$EF2B,$16FC,$8D24,$4016),  {+9.2498149880244712949E+6}
          ($8D43,$C678,$81BB,$C813,$4013),  {+1.6390242166871469602E+6}
          ($D9B5,$7AB8,$B435,$C9B0,$4010),  {+2.0653081576412250327E+5}
          ($A231,$1063,$6764,$9441,$400D),  {+1.8976701935302889156E+4}
          ($059E,$8F9A,$3482,$A022,$4009),  {+1.2810689099125594799E+3}
          ($35CD,$8C87,$6555,$FC3E,$4004),  {+6.3060933434202345361E+1}
          ($950B,$1B89,$3411,$8D47,$4000),  {+2.2074709097925276381}
          ($3037,$2E58,$56F1,$D5A6,$3FFA),  {+5.2160586946135054274E-2}
          ($653E,$8220,$AD06,$C3D1,$3FF4),  {+7.4699038089154483163E-4}
          ($6259,$720D,$001B,$A486,$3FED)); {+4.9031805734598718626E-6}
type
  TX12 = array[0..12] of extended;

  function lratev({$ifdef CONST}const{$else}var{$endif}num: TX12; z: extended): extended;
  const
    denom: TX12 = ( 0.0,
                    39916800.0,
                    120543840.0,
                    150917976.0,
                    105258076.0,
                    45995730.0,
                    13339535.0,
                    2637558.0,
                    357423.0,
                    32670.0,
                    1925.0,
                    66.0,
                    1.0);
  var
    s1,s2: extended;
    i: integer;
  begin
    if abs(z)<=1.0 then begin
      s1 := num[12];
      s2 := denom[12];
      for i:=11 downto 0 do begin
        s1 := s1*z + num[i];
        s2 := s2*z + denom[i];
      end;
    end
    else begin
      z  := 1.0/z;
      s1 := num[0];
      s2 := denom[0];
      for i:=1 to 12 do begin
        s1 := s1*z + num[i];
        s2 := s2*z + denom[i];
      end;
    end;
    lratev := s1/s2;
  end;

begin
  {Ref: Boost [19], lanczos.hpp, struct lanczos13}
  if egscale then lanczos := lratev(TX12(lseh), x)
  else lanczos := lratev(TX12(lsh), x);
end;


{---------------------------------------------------------------------------}
function stirf(x: extended): extended;
  {-Stirling's formula for the gamma function for x > 13.0}
  { gamma(x) = sqrt(2*Pi)*x^(x-0.5)*exp(-x)*(1 + 1/x * P(1/x))}
const
  {13 <= x <= 1024, relative peak error = 9.44E-21, relative error spread = 8.8e-4}
  SPHex: array[0..8] of THexExtW = (
           ($a1d5,$aaaa,$aaaa,$aaaa,$3ffb),  { 8.333333333333331800504E-2}
           ($c3c9,$906e,$38e3,$e38e,$3ff6),  { 3.472222222230075327854E-3}
           ($3a1c,$5ac8,$3478,$afb9,$bff6),  {-2.681327161876304418288E-3}
           ($bef3,$7023,$6a08,$f09e,$bff2),  {-2.294719747873185405699E-4}
           ($30b7,$1a21,$98b2,$cd87,$3ff4),  { 7.840334842744753003862E-4}
           ($5704,$1a39,$b11d,$9293,$3ff1),  { 6.989332260623193171870E-5}
           ($ba6f,$7c59,$5e47,$9bfb,$bff4),  {-5.950237554056330156018E-4}
           ($c395,$0295,$4443,$c64b,$bfef),  {-2.363848809501759061727E-5}
           ($6ede,$69f7,$54e3,$bb5d,$3ff4)); { 7.147391378143610789273E-4}
const
  {x > 1024, rational coefficients from the analytical expansion  }
  { exp(sum(i=1,20,x^(2*i-1)*bernfrac(2*i)/(2*i)/(2*i-1)))+O(x^7) }
  { = 1 + 1/12*x + 1/288*x^2 - 139/51840*x^3 - 571/2488320*x^4    }
  {     + 163879/209018880*x^5 + 5246819/75246796800*x^6 + O(x^7) }
  SAHex: array[0..5] of THexExtW = (
           ($aaab,$aaaa,$aaaa,$aaaa,$3ffb),  { 8.33333333333333333333E-2} {1/12}
           ($e38e,$8e38,$38e3,$e38e,$3ff6),  { 3.47222222222222222222E-3} {1/288}
           ($3df2,$d5a6,$3476,$afb9,$bff6),  {-2.68132716049382716049E-3} {-139/51840}
           ($cab1,$fd42,$7232,$f09e,$bff2),  {-2.29472093621399176955E-4} {-571/2488320}
           ($20e4,$a796,$fb43,$cd87,$3ff4),  { 7.84039221720066627474E-4} {163879/209018880}
           ($c3f2,$ce01,$0241,$923b,$3ff1)); { 6.97281375836585777429E-5} {5246819/75246796800}
const
  XMAX = 1500.0; {max. argument for direct power}
var
  SP: array[0..8] of extended absolute SPHex;
  SA: array[0..5] of extended absolute SAHex;
var
  v,w,y: extended;
begin
  {Ref: Cephes [7], function stirf in file ldouble/gammal.c}
  w := 1.0/x;
  if x > 1024.0 then w := 1.0 + w*PolEvalX(w, SA, 6)
  else w := 1.0 + w*PolEvalX(w, SP, 9);
  if x > XMAX then begin
    {Avoid overflow in power}
    y := exp(-x);
    v := power(x,0.5*x-0.25);
    y := v*(v*y);
  end
  else begin
    y := exp(x);
    y := power(x, x-0.5)/y;
  end;
  stirf := Sqrt_TwoPi*y*w;
end;


{---------------------------------------------------------------------------}
function sfc_signgamma(x: extended): extended;
  {-Return sign(gamma(x)), useless for 0 or negative integer}
begin
  if IsNanOrInf(x) or (x>0.0) or (frac(0.5*floorx(x))=0.0) then sfc_signgamma := 1.0
  else sfc_signgamma := -1.0;
end;


{---------------------------------------------------------------------------}
function sfc_gaminv_small(x: extended): extended;
  {- 1/gamma(x) for small argument values, |x| < 0.03125}
const
  {1/gamma(x) = x*P(x), 0 < x < 0.03125, peak relative error 4.2e-23}
  SPHex: array[0..8] of THexExtW = (
           ($0000,$0000,$0000,$8000,$3fff),   { 1.000000000000000000000E+0}
           ($c7a9,$7db0,$67e3,$93c4,$3ffe),   { 5.772156649015328608253E-1}
           ($7bf6,$57d1,$a013,$a7e7,$bffe),   {-6.558780715202540684668E-1}
           ($f183,$126b,$f47d,$ac0a,$bffa),   {-4.200263503403344054473E-2}
           ($6b8d,$7515,$1905,$aa89,$3ffc),   { 1.665386113720805206758E-1}
           ($10b0,$ec17,$87dc,$acd7,$bffa),   {-4.219773360705915470089E-2}
           ($9225,$dfef,$b0e9,$9da5,$bff8),   {-9.622023360406271645744E-3}
           ($fe9a,$ceb4,$c74e,$ec9a,$3ff7),   { 7.220599478036909672331E-3}
           ($baeb,$d6d3,$25e5,$9c7e,$bff5));  {-1.193945051381510095614E-3}

  {1/gamma(-x) = x*P(x), 0 < x < 0.03125, peak relative error 5.16e-23}
  SNHex: array[0..8] of THexExtW = (
           ($0000,$0000,$0000,$8000,$bfff),   {-1.000000000000000000000E+0}
           ($c7aa,$7db0,$67e3,$93c4,$3ffe),   { 5.772156649015328608727E-1}
           ($5e26,$57d1,$a013,$a7e7,$3ffe),   { 6.558780715202536547116E-1}
           ($7f64,$1234,$f47d,$ac0a,$bffa),   {-4.200263503402112910504E-2}
           ($7a5b,$d76d,$1905,$aa89,$bffc),   {-1.665386113944413519335E-1}
           ($783f,$41dd,$87d1,$acd7,$bffa),   {-4.219773343731191721664E-2}
           ($2ca1,$18f0,$386f,$9da5,$3ff8),   { 9.621911155035976733706E-3}
           ($989b,$dd68,$c5f1,$ec9c,$3ff7),   { 7.220837261893170325704E-3}
           ($5dd1,$02de,$b9f7,$948d,$3ff5));  { 1.133374167243894382010E-3}
var
  SP: array[0..8] of extended absolute SPHex;
  SN: array[0..8] of extended absolute SNHex;
var
  p: extended;
begin
  {Ref: Cephes [7], function gammal/label small in file ldouble/gammal.c}
  if x=0.0 then sfc_gaminv_small := 0.0
  else begin
    if x<0.0 then begin
      x := -x;
      p := PolEvalX(x, SN, 9);
    end
    else p := PolEvalX(x, SP, 9);
    sfc_gaminv_small := x*p;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_gamma_medium(x: extended): extended;
  {-Return gamma(x), |x| <= 13, x negative integer produces div by 0}
const
  {gamma(x+2) = P(x)/Q(x), 0 <= x <= 1, peak relative error = 1.83e-20}
  PHex: array[0..7] of THexExtW = (
           ($0000,$0000,$0000,$8000,$3fff),   { 1.000000000000000000009E+0}
           ($29cf,$19b3,$16c8,$d67a,$3ffe),   { 8.378004301573126728826E-1}
           ($8d75,$23af,$c8e4,$b9d4,$3ffd),   { 3.629515436640239168939E-1}
           ($9549,$8eb5,$8c3a,$e3f4,$3ffb),   { 1.113062816019361559013E-1}
           ($7f43,$5196,$b166,$c368,$3ff9),   { 2.385363243461108252554E-2}
           ($be6c,$3757,$c717,$861b,$3ff7),   { 4.092666828394035500949E-3}
           ($f5aa,$e82f,$335b,$ee2e,$3ff3),   { 4.542931960608009155600E-4}
           ($434a,$3f22,$2bda,$b0b2,$3ff0));  { 4.212760487471622013093E-5}

  QHex: array[0..8] of THexExtW = (
           ($0000,$0000,$0000,$8000,$3fff),   { 9.999999999999999999908E-1}
           ($e458,$2ec7,$fd57,$d47c,$3ffd),   { 4.150160950588455434583E-1}
           ($75ef,$3ab7,$4ad3,$e5bc,$bffc),   {-2.243510905670329164562E-1}
           ($3295,$3698,$d580,$bdcd,$bffa),   {-4.633887671244534213831E-2}
           ($0417,$7989,$d7bc,$e338,$3ff9),   { 2.773706565840072979165E-2}
           ($296e,$7cb1,$5dfd,$d08f,$bff4),   {-7.955933682494738320586E-4}
           ($beed,$1853,$a691,$a23d,$bff5),   {-1.237799246653152231188E-3}
           ($334b,$c2f0,$a2dd,$f60e,$3ff2),   { 2.346584059160635244282E-4}
           ($5473,$2de8,$1268,$ea67,$bfee));  {-1.397148517476170440917E-5}
var
  PX: array[0..7] of extended absolute PHex;
  QX: array[0..8] of extended absolute QHex;
var
  z: extended;
begin
  {Ref: Cephes [7], function gammal in file ldouble/gammal.c}
  {-13 <= x <= 13. Use recurrence formula to bring argument to [2,3)}
  z := 1.0;
  while x >= 3.0 do begin
    x := x-1.0;
    z := z*x;
  end;
  while x < -0.03125 do begin
    z := z/x;
    x := x+1.0;
  end;
  {Here -0.03125 <= x < 3}
  if x <= 0.03125 then begin
    {argument near a pole, use approximation of 1/gamma}
     sfc_gamma_medium := z/sfc_gaminv_small(x);
  end
  else begin
    {finish reduction to [2,3)}
    while x < 2.0 do begin
      z := z/x;
      x := x+1.0;
    end;
    if x=2.0 then  sfc_gamma_medium := z
    else begin
      x := x-2.0;
       sfc_gamma_medium := z * PolEvalX(x,PX,8) / PolEvalX(x,QX,9);
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_lngcorr(x: extended): extended;
  {-Return lngamma correction term lngamma(x) - ((x-0.5)*ln(x) - x + ln(sqrt(2*Pi)), x>=8}
const
  { ln gamma(x) = (x-0.5)*ln(x) - x + ln(sqrt(2*Pi) + 1/x * A(1/x^2) }
  { x >= 8, peak relative error 1.51e-21}
  AHex: array[0..6] of THexExtW = (
          ($9fcc,$aaaa,$aaaa,$aaaa,$3ffb),   { 8.333333333333331447505E-2}
          ($4d88,$03a8,$60b6,$b60b,$bff6),   {-2.777777777750349603440E-3}
          ($f8f2,$30e5,$0092,$d00d,$3ff4),   { 7.936507795855070755671E-4}
          ($8b20,$9fce,$844e,$9c09,$bff4),   {-5.952345851765688514613E-4}
          ($3bdc,$aad1,$d492,$dc88,$3ff4),   { 8.412723297322498080632E-4}
          ($3d91,$0304,$3da1,$f685,$bff5),   {-1.880801938119376907179E-3}
          ($d984,$cc08,$91c2,$a012,$3ff7));  { 4.885026142432270781165E-3}
var
  PA: array[0..6] of extended absolute AHex;
const
  xbig = 4.294967296e9;
  xmax = 2.47860728199e4930;
begin
  {Ref: Cephes [7], part function lgaml in file ldouble/gammal.c}
  if x < XLGAE then begin
    {this is not really used but return 'correct' result}
    sfc_lngcorr := sfc_lngamma(x) - ((x-0.5)*ln(x) - x + LnSqrt2Pi);
  end
  else if x > xmax then sfc_lngcorr := 0.0
  else if x > xbig then sfc_lngcorr := 1.0/(12.0*x)
  else sfc_lngcorr := PolEvalX(1.0/(x*x),PA,7)/x;
end;


{---------------------------------------------------------------------------}
function sfc_gamma(x: extended): extended;
  {-Return gamma(x), x <= MAXGAMX; NAN/RTE if x is a non-positive integer}
var
  i: integer;
  q,z: extended;
begin
  {Ref: Cephes [7], function gammal in file ldouble/gammal.c}
  if x=PosInf_x then begin
    sfc_gamma := x;
    exit;
  end;

  if (x<=0.0) and (frac(x)=0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_gamma := NaN_x;
    exit;
  end;

  if x>MAXGAMX then begin
    sfc_gamma := PosInf_x;
    exit;
  end;

  q := abs(x);
  if q <= 13.0 then sfc_gamma := sfc_gamma_medium(x)
  else begin
    if x < 0.0 then begin
      if q <= MAXGAMX then begin
        i := trunc(q);
        z := q-i;
        if z > 0.5 then z := q - (i+1);
        z := abs(q * sinPi(z)) * stirf(q);
        if z <= Pi/MaxExtended then begin
          sfc_gamma := copysign(PosInf_x, sfc_signgamma(x));
          exit;
        end;
        sfc_gamma := sfc_signgamma(x)*Pi/z;
      end
      else begin
        {Use lngamma, there are only very rare 'non-zero' cases, most values}
        {underflow. gamma(-1756-0.5^24) = -2.75052000244830115684312e-4930}
        z := exp(sfc_lngammas(x,i));
        sfc_gamma := z*i;
      end;
    end
    else sfc_gamma := stirf(x);
  end
end;


{---------------------------------------------------------------------------}
function lngamma_small(x,xm1,xm2: extended; useln1p: boolean): extended;
  {-Return ln(gamma(x), 0<=x<8; xm1=x-1, xm2=x-2 are supplied for increased precision}
  { useln1p should be true if xm1 is supposed to be more accurate than x-1}
var
  r,t,z: extended;
const
  Y20: THexExtW = ($0,$0,$6000,$A2C7,$3FFC); {0.158963680267333984375 = 333371/2097152}
  P20: array[0..6] of extended = (
         -0.180355685678449379109e-1,
          0.25126649619989678683e-1,
          0.494103151567532234274e-1,
          0.172491608709613993966e-1,
         -0.259453563205438108893e-3,
         -0.541009869215204396339e-3,
         -0.324588649825948492091e-4);
  Q20: array[0..7] of extended = (
          1.0,
          0.196202987197795200688e1,
          0.148019669424231326694e1,
          0.541391432071720958364e0,
          0.988504251128010129477e-1,
          0.82130967464889339326e-2,
          0.224936291922115757597e-3,
         -0.223352763208617092964e-6);
const
  Y10: THexExtW = ($0,$0,$1000,$8735,$3FFE);  {0.52815341949462890625 = 1107618/2097152}
  P10: array[0..6] of extended = (
          0.490622454069039543534e-1,
         -0.969117530159521214579e-1,
         -0.414983358359495381969e0,
         -0.406567124211938417342e0,
         -0.158413586390692192217e0,
         -0.240149820648571559892e-1,
         -0.100346687696279557415e-2);
  Q10: array[0..6] of extended = (
          1.0,
          0.302349829846463038743e1,
          0.348739585360723852576e1,
          0.191415588274426679201e1,
          0.507137738614363510846e0,
          0.577039722690451849648e-1,
          0.195768102601107189171e-2);
const
  Y15: THexExtW = ($0,$0,$D000,$E76E,$3FFD);  {0.452017307281494140625 = 947949/2097152}
  P15: array[0..5] of extended = (
         -0.292329721830270012337e-1,
          0.144216267757192309184e0,
         -0.142440390738631274135e0,
          0.542809694055053558157e-1,
         -0.850535976868336437746e-2,
          0.431171342679297331241e-3);
  Q15: array[0..6] of extended = (
          1.0,
         -0.150169356054485044494e1,
          0.846973248876495016101e0,
         -0.220095151814995745555e0,
          0.25582797155975869989e-1,
         -0.100666795539143372762e-2,
         -0.827193521891290553639e-6);
begin
  {Based on \boost\math\special_functions\detail\lgamma_small.hpp [19]}
  {Copyright John Maddock 2006, see 3rdparty.ama for Boost license}
  if x<eps_x then begin
    lngamma_small := -ln(x);
    exit;
  end;
  if (xm1=0.0) or (xm2=0.0) then begin
    lngamma_small := 0.0;
    exit;
  end;
  {The Boost rational approximations are optimized for low absolute error.}
  {As long as their absolute error is small compared to the constants Yxx }
  {then any rounding errors in their computation will get wiped out.      }
  if x>2.0 then begin
    {Use recurrence formula to bring argument to [2,3)}
    if x>=3 then begin
      z := 1.0;
      repeat
        x := x-1.0;
        z := z*x;
      until x<3.0;
      xm2 := x-2.0;
      z := ln(z);
    end
    else z := 0.0;
    {Use the following form: lngamma(x) = (x-2)*(x+1)*(Y20 + R(x-2))}
    r := PolEvalX(xm2, P20, 7) / PolEvalX(xm2, Q20, 8);
    t := xm2*(x+1.0);
    lngamma_small := z + (t*extended(Y20) + t*r);
  end
  else begin
    if x<1.0 then begin
      {bring argument into [1,2)}
      if useln1p then z := -ln1p(xm1)
      else z := -ln(x);
      xm2 := xm1;
      xm1 := x;
      x   := x+1.0;
    end
    else z := 0.0;
    {Two approximations, one for x in [1.0,1.5] and one for x in (1.5,2]}
    if x<=1.5 then begin
      {Use the following form: lngamma(x) = (x-1)*(x-2)*(Y10 + R(x-1))}
      r := PolEvalX(xm1, P10, 7) / PolEvalX(xm1, Q10, 7);
      t := xm1*xm2;
      lngamma_small := z + (t*extended(Y10) + t*r);
    end
    else begin
      {Use the following form: lngamma(x) = (2-x)*(1-x)*(Y15 + R(2-x))}
      r := PolEvalX(-xm2, P15, 6) / PolEvalX(-xm2, Q15, 7);
      t := xm1*xm2;
      lngamma_small := z + (t*extended(Y15) + t*r);
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_lngamma(x: extended): extended;
  {-Return ln(|gamma(x)|), |x| <= MAXLGM, NAN/RTE if x is a non-positive integer}
var
  t: extended;
begin
  if x<=0.0 then begin
    {avoid infinite loop if x is (negative) NAN}
    if IsNaN(x) then sfc_lngamma := NaN_x
    else if frac(x)=0.0 then begin
      {$ifopt R+}
        if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      {$endif}
      sfc_lngamma := NaN_x;
    end
    else begin
      {use reflection formula}
      t := x*sinPi(x);
      sfc_lngamma := ln(abs(Pi/t)) - sfc_lngamma(abs(x));
    end;
    exit;
  end;
  if x <= XLGAE then begin
    {Note: will generate error for x=0}
    sfc_lngamma := lngamma_small(x,x-1.0,x-2.0,false)
  end
  else begin
    {Stirling approximation for x > 8}
    sfc_lngamma := (x-0.5)*ln(x) - x + LnSqrt2Pi + sfc_lngcorr(x);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_lngammas(x: extended; var s: integer): extended;
  {-Return ln(|gamma(x)|), |x| <= MAXLGM, s=-1,1 is the sign of gamma}
begin
  sfc_lngammas := sfc_lngamma(x);
  if (x>0.0) or (frac(0.5*floorx(x))=0.0) then s := 1 else s := -1;
end;


{---------------------------------------------------------------------------}
function sfc_lngamma1p(x: extended): extended;
  {-Return ln(|gamma(1+x)|) with increased accuracy for x near 0}
var
  t: extended;
begin
  t := 1.0+x;
  if (t < 0.0) or (t > XLGAE) then sfc_lngamma1p := sfc_lngamma(t)
  else sfc_lngamma1p := lngamma_small(t,x,x-1.0,true)
end;


{---------------------------------------------------------------------------}
function sfc_gamma1pm1(x: extended): extended;
  {-Return gamma(1+x)-1 with increased accuracy for x near 0}
begin
  if abs(x)<=eps_x then sfc_gamma1pm1 := -EulerGamma*x
  else if (x<-0.5) or (x>2.0) then sfc_gamma1pm1 := sfc_gamma(1.0+x)-1.0
  else if x>0.0 then sfc_gamma1pm1 := expm1(lngamma_small(x+1.0,x,x-1.0,true))
  else sfc_gamma1pm1 := expm1(lngamma_small(x+2.0,x+1.0,x,false) - ln1p(x));
end;


{---------------------------------------------------------------------------}
function sfc_rgamma(x: extended): extended;
  {-Return the reciprocal gamma function rgamma = 1/gamma(x)}
begin
  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_rgamma := 0.0
    else sfc_rgamma := NaN_x;
  end
  else if abs(x)<0.03125 then sfc_rgamma := sfc_gaminv_small(x)
  else if (x<0.0) and (frac(x)=0.0) then sfc_rgamma := 0.0
  else begin
    {
    if (x<-0.5) or (x>XLGAE) then sfc_rgamma := sfc_signgamma(x)*exp(-sfc_lngamma(x))
    else sfc_rgamma := 1.0/sfc_gamma_medium(x);
    }
    if (x<-0.5) or (x>=MAXGAMX) then sfc_rgamma := sfc_signgamma(x)*exp(-sfc_lngamma(x))
    else if x <= 13.0 then sfc_rgamma := 1.0/sfc_gamma_medium(x)
    else sfc_rgamma := 1.0/stirf(x);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_gstar(x: extended): extended;
  {-Return Temme's gstar(x) = gamma(x)/(sqrt(2*Pi)*x^(x-0.5)*exp(-x)), x>0.}
  { For large x the asymptotic expansion is gstar(x) = 1 + 1/12x + O(1/x^2)}
var
  a,g: extended;
const
  {Chebyshev approximations calculated with Maple V R4 with Digits :=30;}
  {Coefficients for T(0,.) doubled, HexExtended with t_rcalc command xh.}
  {Definition: gstar := x-> GAMMA(x)/(sqrt(2*Pi)*x^(x-0.5)*exp(-x));}

  {Range 1<=x<= 3; y := t -> t+2;}
  {chebyshev(y(t)^2*(gstar(y(t))-1-1/(12*y(t))), t=-1..1, 0.5e-20);}
  gs1n = 30;
  gs1h : array[0..gs1n-1] of THexExtW = (
           ($26A1,$14BA,$EC57,$83AB,$3FF7),  { 4.0182975773513416011E-3}
           ($8A43,$A98D,$2BD2,$B827,$3FF4),  { 7.0248799299004439952E-4}
           ($81EF,$77ED,$7E07,$A731,$BFF2),  {-1.5944798403032040299E-4}
           ($8F60,$F4D0,$DFFD,$900D,$3FF0),  { 3.4345197462305170773E-5}
           ($C34A,$42D7,$2A10,$EB6B,$BFED),  {-7.0160213288181423859E-6}
           ($7A47,$6EED,$A8E0,$B53C,$3FEB),  { 1.3503205169277659442E-6}
           ($7D24,$93CC,$D582,$8133,$BFE9),  {-2.4065836641022231323E-7}
           ($AFE6,$97BF,$5311,$A364,$3FE6),  { 3.8042639505142093235E-8}
           ($63DE,$2861,$E471,$A014,$BFE3),  {-4.6589880584690519361E-9}
           ($2FEC,$9CDC,$9ED8,$94B3,$3FDE),  { 1.3524335604397218593E-10}
           ($543F,$B1E8,$647F,$C44B,$3FDE),  { 1.7852880977454193496E-10}
           ($B937,$7641,$3927,$C00A,$BFDD),  {-8.7329651525165672802E-11}
           ($168B,$D18D,$7F0B,$84E5,$3FDC),  { 3.0217158796552201555E-11}
           ($1576,$BF4D,$8C8F,$A133,$BFDA),  {-9.1632366291192802894E-12}
           ($F8A3,$ABB3,$AEA7,$B66A,$3FD8),  { 2.5922976124555003060E-12}
           ($9848,$57E9,$3BE2,$C5E9,$BFD6),  {-7.0312136560940930257E-13}
           ($B155,$A152,$D1F5,$D0E3,$3FD4),  { 1.8553152123744796025E-13}
           ($DE6F,$8669,$0937,$D861,$BFD2),  {-4.8045799980445695572E-14}
           ($9337,$6A1D,$6404,$DD3A,$3FD0),  { 1.2280625885209937902E-14}
           ($91A1,$6ACB,$2EC2,$E020,$BFCE),  {-3.1103690941234212527E-15}
           ($4020,$F48D,$C5F0,$E19E,$3FCC),  { 7.8277734240176922378E-16}
           ($9D0A,$0C51,$259C,$E224,$BFCA),  {-1.9614622328820518705E-16}
           ($6E3B,$3B7E,$7731,$E205,$3FC8),  { 4.9010567735366476461E-17}
           ($8153,$BE84,$AC06,$E183,$BFC6),  {-1.2225157066341578107E-17}
           ($5EFA,$2FDE,$EA16,$E0CE,$3FC4),  { 3.0467200401436511414E-18}
           ($6370,$29B3,$9E32,$E013,$BFC2),  {-7.5920116218493385258E-19}
           ($6F91,$69A3,$68EE,$DF41,$3FC0),  { 1.8910477040223097493E-19}
           ($80CF,$9130,$ABE5,$DF35,$BFBE),  {-4.7266482810237095702E-20}
           ($6062,$3806,$EC04,$DCC2,$3FBC),  { 1.1687011937049585628E-20}
           ($01FE,$0907,$EE5B,$E372,$BFBA)); {-3.0102617821884302625E-21}
var
  gs1a: array[0..gs1n-1] of extended absolute gs1h;

const
  gs2n = 31; {chebyshev(gstar(t), t=3..8, 0.5e-20);}
  gs2h : array[0..gs2n-1] of THexExtW = (
           ($78D4,$FA06,$AD41,$8231,$4000),  { 2.0342820305159885945}
           ($D94B,$BBBA,$AB78,$87C7,$BFF8),  {-8.2873510863767290591E-3}
           ($D710,$9AB3,$309C,$831F,$3FF6),  { 2.0007604294772196608E-3}
           ($664B,$4102,$BD79,$FCED,$BFF3),  {-4.8242315747915459550E-4}
           ($6E6D,$ABF2,$018C,$F3A7,$3FF1),  { 1.1618250245736783082E-4}
           ($BAD0,$512E,$362E,$EA73,$BFEF),  {-2.7948623357453920514E-5}
           ($427A,$FE0E,$B94A,$E15A,$3FED),  { 6.7160841776690551181E-6}
           ($7FC4,$E615,$27FF,$D865,$BFEB),  {-1.6122694432402287168E-6}
           ($B1F8,$8BDA,$4FF3,$CF99,$3FE9),  { 3.8668303975989479479E-7}
           ($9AAE,$587F,$2775,$C6FD,$BFE7),  {-9.2661419574693789809E-8}
           ($0655,$F513,$CBA1,$BE95,$3FE5),  { 2.2187030226642767308E-8}
           ($418D,$BD3D,$83AE,$B667,$BFE3),  {-5.3086653662215709374E-9}
           ($51AD,$FD6D,$C872,$AE75,$3FE1),  { 1.2693642192007070286E-9}
           ($2E7A,$45B5,$4F47,$A6C3,$BFDF),  {-3.0333999974117580609E-10}
           ($FEA3,$E5E7,$175E,$9F52,$3FDD),  { 7.2450652218555779523E-11}
           ($7398,$B75A,$78D2,$9823,$BFDB),  {-1.7296152050716179453E-11}
           ($957B,$0418,$34AC,$9138,$3FD9),  { 4.1273879591581087674E-12}
           ($43E9,$5D5B,$8558,$8A90,$BFD7),  {-9.8456023555037793839E-13}
           ($DBD4,$B67F,$2F09,$842C,$3FD5),  { 2.3478568909854971813E-13}
           ($DA03,$7484,$1F3F,$FC15,$BFD2),  {-5.5973560909513051708E-14}
           ($8336,$2310,$5B9E,$F055,$3FD0),  { 1.3341185335663677403E-14}
           ($3600,$2A12,$8EA1,$E516,$BFCE),  {-3.1792362333608899741E-15}
           ($D32F,$A155,$BBE9,$DA55,$3FCC),  { 7.5750134834532103750E-16}
           ($7F27,$7DFA,$870F,$D00F,$BFCA),  {-1.8046385098946929066E-16}
           ($7E48,$8F57,$48FA,$C640,$3FC8),  { 4.2988857605477843878E-17}
           ($8495,$0C96,$22D2,$BCE4,$BFC6),  {-1.0239810102203542236E-17}
           ($5E48,$E817,$0E96,$B3F7,$3FC4),  { 2.4389814490930843665E-18}
           ($D507,$1A15,$ED88,$AB74,$BFC2),  {-5.8091806323722222819E-19}
           ($1C8E,$8568,$9477,$A359,$3FC0),  { 1.3836276558996084587E-19}
           ($4846,$9E37,$D63A,$9BA0,$BFBE),  {-3.2955567815269497085E-20}
           ($2038,$2762,$8C64,$9446,$3FBC)); { 7.8496438299511273955E-21}
var
  gs2a: array[0..gs2n-1] of extended absolute gs2h;
const
  gstar1: THexExtW = ($F232,$F32F,$D984,$8ACE,$3FFF);  {1.0844375514192275466}
begin
  if x>=1.79e8 then begin
    {very large x, use first terms of Stirling formula}
    a := 1.0/(12.0*x);
    sfc_gstar := 1.0 + a*(1.0 + 0.5*a);
  end
  else if x>8.0 then begin
    {medium range, recycle sfc_lngcorr}
    sfc_gstar := exp(sfc_lngcorr(x));
  end
  else if x>=3.0 then begin
    {3 <= x < 8}
    sfc_gstar := CSEvalX((2.0*x-11.0)/5.0, gs2a, gs2n);
  end
  else if x>=1.0 then begin
    {1 <= x < 3}
    a :=  CSEvalX(x-2.0, gs1a, gs1n);
    sfc_gstar := a/sqr(x) + 1.0 + 1.0/(12.0*x);
  end
  else if x>=MinExtended then begin
    {0 < x < 1: Use Temme's [26] recursion formula for gstar(x+1),}
    {but avoid recursive call if x+1 = 1 within machine precision.}
    g := 1.0 + x;
    if g=1.0 then g := extended(gstar1)
    else g := sfc_gstar(g);
    a := 1.0 + 1.0/x;
    g := g*sqrt(a);
    a := ln(a)*x - 1.0;
    sfc_gstar := g*exp(a);
  end
  else sfc_gstar := 1.0/(Sqrt_TwoPi*sqrt(x));
end;


{---------------------------------------------------------------------------}
const
  FTabH: array[0..25] of THexExtW = (        {Table of factorials}
           ($0000,$0000,$0000,$8000,$3FFF),  {1.000000000000000000000}
           ($0000,$0000,$0000,$8000,$3FFF),  {1.000000000000000000000}
           ($0000,$0000,$0000,$8000,$4000),  {2.000000000000000000000}
           ($0000,$0000,$0000,$C000,$4001),  {6.000000000000000000000}
           ($0000,$0000,$0000,$C000,$4003),  {2.400000000000000000000E+1}
           ($0000,$0000,$0000,$F000,$4005),  {1.200000000000000000000E+2}
           ($0000,$0000,$0000,$B400,$4008),  {7.200000000000000000000E+2}
           ($0000,$0000,$0000,$9D80,$400B),  {5.040000000000000000000E+3}
           ($0000,$0000,$0000,$9D80,$400E),  {4.032000000000000000000E+4}
           ($0000,$0000,$0000,$B130,$4011),  {3.628800000000000000000E+5}
           ($0000,$0000,$0000,$DD7C,$4014),  {3.628800000000000000000E+6}
           ($0000,$0000,$4000,$9845,$4018),  {3.991680000000000000000E+7}
           ($0000,$0000,$E000,$E467,$401B),  {4.790016000000000000000E+8}
           ($0000,$0000,$6600,$B994,$401F),  {6.227020800000000000000E+9}
           ($0000,$0000,$D940,$A261,$4023),  {8.717829120000000000000E+10}
           ($0000,$0000,$BBAC,$983B,$4027),  {1.307674368000000000000E+12}
           ($0000,$0000,$BBAC,$983B,$402B),  {2.092278988800000000000E+13}
           ($0000,$C000,$7766,$A1BF,$402F),  {3.556874280960000000000E+14}
           ($0000,$9800,$6653,$B5F7,$4033),  {6.402373705728000000000E+15}
           ($0000,$4480,$C983,$D815,$4037),  {1.216451004088320000000E+17}
           ($0000,$0AD0,$9DF2,$870D,$403C),  {2.432902008176640000000E+18}
           ($0000,$AE31,$DF4D,$B141,$4040),  {5.109094217170944000000E+19}
           ($6000,$CF83,$930A,$F3BA,$4044),  {1.124000727777607680000E+21}
           ($6D00,$C526,$19AF,$AF2E,$4049),  {2.585201673888497664000E+22}
           ($D1C0,$D3DC,$9343,$8362,$404E),  {6.204484017332394393600E+23}
           ($07BC,$FB09,$0619,$CD4A,$4052)); {1.551121004333098598400E+25}
var
  FacTab: array[0..25] of extended absolute FTabH;


{---------------------------------------------------------------------------}
function sfc_fac(n: integer): extended;
  {-Return the factorial n!, n<MAXGAMX-1; INF if n<0}
begin
  if n>25 then sfc_fac := stirf(n+1)
  else begin
    if n<0 then sfc_fac := PosInf_x
    else sfc_fac := FacTab[n];
  end;
end;


{---------------------------------------------------------------------------}
function sfc_lnfac(n: longint): extended;
  {-Return ln(n!), INF if n<0}
begin
  if n>25 then sfc_lnfac := sfc_lngamma(n+1.0)
  else begin
    if n<0 then sfc_lnfac := PosInf_x
    else sfc_lnfac := ln(FacTab[n]);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_dfac(n: integer): extended;
  {-Return the double factorial n!!, n<=3209; INF for even n<0}
var
  k: integer;
  t: extended;
begin
  if n<0 then begin
    if odd(n) then begin
      if n=-1 then sfc_dfac := 1
      else begin
        {(-2k-1)!! = (-1)^k/(2k-1)!!}
        t := 1.0/sfc_dfac(-n-2);
        if n and 2 = 0 then sfc_dfac := -t
        else sfc_dfac := t;
      end;
    end
    else sfc_dfac := PosInf_x;
  end
  else if n>3209 then sfc_dfac := PosInf_x
  else if odd(n) then begin
    { (2k+1)!! = (2k+1)!/(2^k * k!) }
    if n <= 25 then begin
      k := n div 2;
      sfc_dfac := FacTab[n]/ldexp(FacTab[k],k);
    end
    else begin
      {Here n >= 27 and the argument of stirf is >= 14.5}
      n := succ(n) div 2;
      t := ldexp(stirf(n+0.5), n);
      sfc_dfac := int(t/SqrtPi+0.5);
    end;
  end
  else begin
    { n!! = 2^k * k!  with k = n div 2}
    n := n div 2;
    sfc_dfac := ldexp(sfc_fac(n),n);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_binomial(n,k: integer; dbl: boolean): extended;
  {-Return the binomial coefficient 'n choose k'; set dbl=true, if used in SpecFun}
var
  t,a: extended;
  i,m: integer;
  function lnf0(m: integer): extended;
    {-Return quick and dirty approximation of ln(m!)}
  begin
    lnf0 := (0.5+m)*ln(1.0+m)-m;
  end;
begin
  (*
  Accuracy of sfc_binomial, random values 0<=k<=n, computed with t_binom
  Range (n)      Samples   RMS           Peak
      0.. 1000   20000     1.2087E-19    5.9631E-19  for (820,415)
   1000.. 1700   20000     1.4733E-19    6.5052E-19  for (1678,331)
   1700.. 8000   20000     1.4177E-18    7.4801E-18  for (7185,4779)
   8000..16000   10000     2.0377E-16    1.2166E-15  for (14557,6857)
  16000..32000   10000     1.8608E-16    1.2806E-15  for (18717,4636)
  *)

  {If n<0 and k>=0, we use the identity 1.2.6 (17) from D.E. Knuth: The art}
  {of computer programming, Vol. 1, Fundamental algorithms, 3rd ed., 1997.}
  {[*]   binomial(n,k) = (-1)^k * binomial(k-n-1,k)}

  if (k=0) or (k=n) then sfc_binomial := 1.0
  else if (k=1) or (k=pred(n)) then sfc_binomial := n
  else if k<0 then begin
    if (n>=0) or (n<k) then sfc_binomial := 0.0
    else begin
      {Here k <= n < 0:  We use binomial(n,k) = binomial(n,n-k) which is}
      {valid for all integer k, and then use [*] since n<0 and n-k >= 0:}
      {binomial(n,k) = (-1)^(n-k)*binomial(n-k-n-1,n-k)}
      i := n-k;
      t := sfc_binomial(-succ(k),i,dbl);
      if (t<>0.0) and odd(i) then sfc_binomial := -t
      else sfc_binomial := t;
    end;
  end
  else if n<0 then begin
    {Here n < 0 and k >= 0. Apply Knuth's [*] identity}
    t := sfc_binomial(k-n-1,k,dbl);
    if (t<>0.0) and odd(k) then sfc_binomial := -t
    else sfc_binomial := t;
  end
  else if k>n then sfc_binomial := 0.0
  else begin
    {here n > 2 and  1 < k < n-1}
    if n<MAXGAMX-1 then begin
      t := sfc_fac(n);
      t := t/sfc_fac(n-k);
      t := t/sfc_fac(k);
    end
    else begin
      {if k>n/2 use symmetry to reduce k}
      if k>n-k then k := n-k;
      {all (n,k) with n <= m will NOT overflow,}
      {this saves three logarithm calculations.}
      if dbl then begin
        a := 750;
        m := 1029;
      end
      else begin
        a := 11500;
        m := 16391;
      end;
      if n<=m then t:=0
      else begin
        {Calculate rough estimate t}
        t := lnf0(k);
        t := t + lnf0(n-k);
        t := lnf0(n)-t;
      end;
      if t>a then begin
        {don't waste time (especially for double), if result is INF}
        t := PosInf_x
      end
      else if (t<11300) and (k<=4000) then begin
        {Compute n*(n-1).../k! if k is 'small' and no risk of overflow}
        t := n;
        inc(n);
        for i:=2 to k do t := t*((n-i)/i);
      end
      else begin
        {Here n > 8000 and k > 4000, or a very large result > 1.7e4777.}
        {Use lnbeta function (increased inaccuracy for 'large' values)}
        t := sfc_lnbeta(k,n-k+1);
        t := k*exp(t);
        if t<MinExtended then begin
          {for t > 0.25*MinExtended inversion will not overflow}
          if 4.0*t<MinExtended then t := PosInf_x
          else t := 1.0/t;
        end
        else t := 1.0/t;
      end;
    end;
    {round to nearest integer value}
    sfc_binomial := int(t+0.5);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_gdr_pos(x,d: extended): extended;
  {-Return gamma(x)/gamma(x+d), x>0, x+d > 0, accurate even for |d| << |x|}
var
  xgh,res: extended;
begin
  {$ifdef debug}
    if (x <= 0.0) and (x+d <= 0.0) then begin
      {$ifopt R+}
        if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      {$endif}
      sfc_gdr_pos := NaN_x;
      exit;
    end;
  {$endif}
  xgh := x + lanczos_gm05;
  if abs(d) < 10.0 then begin
    res := ln1p(d/xgh);
    res := exp((0.5-x)*res);
  end
  else begin
    res := power(xgh/(xgh+d), x-0.5);
  end;
  res := res*power(exp(1.0)/(xgh + d), d);
  sfc_gdr_pos := lanczos(x,false)/lanczos(x+d,false)*res;
end;


{---------------------------------------------------------------------------}
function sfc_gamma_delta_ratio(x,d: extended): extended;
  {-Return gamma(x)/gamma(x+d), accurate even for |d| << |x|}
var
  s,t,y: extended;
begin
  if IsNanOrInf(x) or IsNanOrInf(d) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_gamma_delta_ratio := NaN_x;
    exit;
  end;
  y := x+d;
  if (x>0.0) and (y>0.0) then sfc_gamma_delta_ratio := sfc_gdr_pos(x,d)
  else begin
    if (y<=0.0) and (frac(y)=0.0) then begin
      {y=x+d is a non-positive integer}
      if (x>0.0) or (frac(x)<>0.0) then begin
        {gamma(x+d) = inf, gamma(x)<>0, return zero}
        sfc_gamma_delta_ratio := 0.0;
        exit;
      end;
      {both x,y non-positive integers, no need to calculate sin for reflection}
      t := sfc_gdr_pos(1.0-x,-d);
      if frac(0.5*x)<>frac(0.5*y) then t := -t;
      sfc_gamma_delta_ratio := 1.0/t;
      exit;
    end;
    if (x<0.0) and (y<0.0) then begin
      {Here both x and y are negative, use reflection formula for gammas}
      {gamma(x)/gamma(x+d) = sinPi(x+d)/sinP(x)/ [gamma(1-x)/gamma(1-x-d]}
      t := sfc_gdr_pos(1.0-x,-d);
      t := sinPi(x)*t;
      y := sinPi(y);
      sfc_gamma_delta_ratio := y/t;
    end
    else begin
      {Remaing cases: use lngamma for both arguments}
      s := sfc_signgamma(x)*sfc_signgamma(y);
      y := sfc_lngamma(y);
      t := sfc_lngamma(x);
      t := exp(t-y);
      sfc_gamma_delta_ratio := s*t;
    end
  end;
end;


{---------------------------------------------------------------------------}
function sfc_gamma_ratio(x,y: extended): extended;
  {-Return gamma(x)/gamma(y)}
begin
  sfc_gamma_ratio := sfc_gamma_delta_ratio(x,y-x)
end;


{---------------------------------------------------------------------------}
function sfc_pochhammer(a,x: extended): extended;
  {-Return the Pochhammer symbol (a)_x = gamma(a+x)/gamma(a).}
  { Accuracy is reduced if x or x+a are near negative integers.}
var
  t: extended;
  it,ia: boolean;
  i: integer;
begin
  if IsNanOrInf(x) or IsNanOrInf(a) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_pochhammer := NaN_x;
    exit;
  end;
  if x=0.0 then sfc_pochhammer := 1.0
  else begin
    t := a+x;
    ia := (a<=0.0) and (frac(a)=0.0);
    it := (t<=0.0) and (frac(t)=0.0);
    if it or ia then begin
      if it and ia then begin
        {both a and a+x are negative integers}
        t := sfc_gamma_delta_ratio(1-a,-x);
        if frac(0.5*x)<>0 then t := -t;
        sfc_pochhammer := t;
      end
      else if ia then sfc_pochhammer := 0.0
      else begin
        {$ifopt R+}
          if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
        {$endif}
        sfc_pochhammer := NaN_x;
      end;
    end
    else begin
      if (frac(x)=0.0) and (x>0.0) and (x<=100.0) then begin
        {x is a small positive integer, use simple multiply loop}
        t := a;
        for i:=1 to trunc(x)-1 do t := t*(a+i);
        sfc_pochhammer := t;
      end
      else begin
        t := sfc_gamma_delta_ratio(a,x);
        if t<>0 then sfc_pochhammer := 1.0/t
        else sfc_pochhammer := PosInf_x;
      end;
    end;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_poch1(a,x: extended): extended;
  {-Return (pochhammer(a,x)-1)/x, psi(a) if x=0; accurate even for small |x|}
const
  sqtbig = 0.111324015423318091e2466;  {=1/sqrt(24*MinExtended)}
  lneps  = -44.3614195558364998;       {=ln(eps_x/2)}
var
  gbern : array[1..21] of extended;
var
  t,b,bp,v,v2,lv,q,poly,res,term: extended;
  nterms,incr,j,k: integer;
begin
  {Ref: W. Fullerton [14] and [20], files dpoch1.f}
  if IsNanOrInf(x) or IsNanOrInf(a) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_poch1 := NaN_x;
    exit;
  end;

  if x=0.0 then begin
    sfc_poch1 := sfc_psi(a);
    exit;
  end;

  t := abs(x);
  q := abs(a);
  if (t > 0.1*q) or (t*ln(maxx(q,2.0)) > 0.1) then begin
    res := sfc_pochhammer(a,x);
    sfc_poch1 := (res-1.0)/x;
    exit;
  end;

  if a  < -0.5 then bp := 1.0-a-x else bp := a;
  if bp < 10.0 then incr := trunc(11.0 - bp) else incr := 0;

  b  := bp + incr;
  v  := b + 0.5*(x-1.0);
  lv := ln(v);
  q  := x*lv;

  if v < sqtbig then begin
    nterms := trunc(-0.5*lneps/lv + 1.0);
    if nterms>20 then begin
      {$ifdef debug}
        sfc_write_debug_str('*** sfc_poch1: nterms too big');
      {$endif}
      nterms := 20;
    end;
    v2  := 1.0/sqr(v);
    res := 0.5*(x+1.0);
    gbern[1] := 1.0;
    gbern[2] := -res/12.0;
    term := v2;
    poly := gbern[2]*term;
    for k:=2 to nterms do begin
      t := 0.0;
      for j:=1 to k do t := t + extended(BoFHex[k-j])*gbern[j];
      gbern[k+1] := -res*t/k;
      term := term * ((2*k-2)-x)*((2*k-1)-x)*v2;
      poly := poly + gbern[k+1]*term;
    end;
  end
  else poly := 0.0;

  poly := (x-1.0)*poly;
  res  := exprel(q)*(lv + q*poly) + poly;

  if incr > 0 then begin
    {Get poch1(bp,x) for small bp from poch1(b,x) with backwards recursion}
    for j:=pred(incr) downto 0 do begin
      t := 1.0/(bp+j);
      res := (res-t)/(x*t+1.0);
    end;
  end;

  if bp=a then sfc_poch1 := res
  else begin
    {Use reflection formula to get poch1(a,x) with a<-0.5 from poch1(bp,x)}
    {q = cot(Pi*b)}
    sincosPi(b,t,q);
    q := q/t;
    t := sinPi(x)/x;
    t := t*q - 2.0*sqr(sinPi(0.5*x))/x;
    sfc_poch1 := t + (1.0 + x*t)*res;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_psi(x: extended): extended;
  {-Return the psi (digamma) function of x, INF if x is a non-positive integer}
const
  BNNHex: array[0..8] of THexExtW = (      {bernreal(2k+2)/(2k+2), k=0..7}
            ($AAAB, $AAAA, $AAAA, $AAAA, $3FFB),   { 8.33333333333333333333333E-2}
            ($8889, $8888, $8888, $8888, $BFF8),   {-8.33333333333333333333333E-3}
            ($8208, $0820, $2082, $8208, $3FF7),   { 3.96825396825396825396825E-3}
            ($8889, $8888, $8888, $8888, $BFF7),   {-4.16666666666666666666667E-3}
            ($3E10, $E0F8, $0F83, $F83E, $3FF7),   { 7.57575757575757575757576E-3}
            ($ACCB, $CACC, $CCAC, $ACCA, $BFF9),   {-2.10927960927960927960928E-2}
            ($AAAB, $AAAA, $AAAA, $AAAA, $3FFB),   { 8.33333333333333333333333E-2}
            ($F2F3, $F2F2, $F2F2, $E2F2, $BFFD),   {-4.43259803921568627450980E-1}
            ($3FCE, $FF37, $FCDC, $C373, $4000));  { 3.05395433027011974380395   }
var
  BNN: array[0..7] of extended absolute BNNHex;

const
  PPHex: array[0..5] of THexExtW = (
           ($5D8F, $71C5, $5E47, $F7B9, $3FFE),  {+0.967672245447621170427444761712   }
           ($93CE, $3DA8, $4992, $9274, $3FFF),  {+1.14417380943934132177443086511    }
           ($6D52, $5F65, $79EF, $F169, $3FFD),  {+0.471507845373433246720039154386   }
           ($9EF2, $DF82, $1865, $A71D, $3FFB),  {+0.815984636391829369377116423465e-1}
           ($D564, $B95D, $BF51, $B443, $3FF7),  {+0.550124017486102827456755050064e-2}
           ($05E4, $B416, $530F, $C9AD, $3FF1)); {+0.961671107604462646458076964138e-4}
  PQHex: array[0..6] of THexExtW = (
           ($0000, $0000, $0000, $8000, $3FFF),  {+1.00000000000000000000000000000    }
           ($A4B3, $CD52, $FAA6, $D1E9, $3FFF),  {+1.63995297569876643247016291140    }
           ($2D1C, $1958, $35D7, $F872, $3FFE),  {+0.970492711080937113643931469693   }
           ($E38A, $0D1E, $6FB5, $84F8, $3FFD),  {+0.259707918978674869994308187160   }
           ($A4B6, $ADD5, $3BC7, $81BF, $3FFA),  {+0.316765151172736832756607905517e-1}
           ($C902, $DFDF, $6A6A, $C67A, $3FF5),  {+0.151426838914381207406214536130e-2}
           ($B01C, $4DC7, $7496, $8E9D, $3FEF)); {+0.170010400090619970545999542634e-4}
const
  x0hHex: THexExtW = ($0000,$0000,$8000,$BB16,$3FFF);  {1.4616241455078125}
  x0lHex: THexExtW = ($A9C9,$F6E1,$6BE3,$8635,$3FEE);  {7.9994605498412626595423257213e-6}
var
  PP : array[0..5] of extended absolute PPHex;
  PQ : array[0..6] of extended absolute PQHex;
  x0h: extended absolute x0hHex;
  x0l: extended absolute x0lHex;
var
  p,q,nz,s,w,y,z: extended;
  i,n: integer;
  subneg: boolean;

begin

  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_psi := PosInf_x
    else sfc_psi := NaN_x;
    exit;
  end;

  {General algorithm layout is as in Cephes [7] double/psi.c. Coefficients }
  {of the asymptotic expansion are calculated with Pari/GP's bernreal, the }
  {Pad approximation polynomials with Maple V's "pade(Psi(x+x0),x,[6,6])".}
  {Conversions to THexExtW form are done with MPArith/t_rcalc's xh command.}

  subneg := false;
  nz := 0.0;

  if x<=0.0 then begin
    q := x;
    p := floorx(q);
    if p=q then begin
      sfc_psi := PosInf_x;
      exit;
    end;
    {Remove the zeros of tan(Pi*x) by subtracting the nearest integer from x}
    nz := q - p;
    if nz<>0.5 then begin
      if nz>0.5 then begin
        p  := p + 1.0;
        nz := q - p;
      end;
      {nz := Pi/tan(Pi*nz)}
      sincosPi(nz,s,z);
      nz := Pi*z/s;
    end
    else nz := 0.0;
    x := 1.0 - x;
    subneg := nz<>0.0;
  end;

  {Check for small positive integers}
  if (x <= 12.0) and (x=floorx(x)) then begin
    y := 0.0;
    n := trunc(x);
    for i:=n-1 downto 1 do y := y+1.0/i;
    y := y - EulerGamma;
  end
  else if abs(x-x0h)<0.2 then begin
    {Pad approximation for x in (x0-0.2, x0+0.2), abs. error < 3e-19}
    {where x0=1.46163214496836234126.., psi(x0)=0}
    x := x-x0h;
    x := x-x0l;
    y := x*(PolEvalX(x,PP,6)/PolEvalX(x,PQ,7));
  end
  else begin
    s := x;
    w := 0.0;
    while s<12.0 do begin
      w := w + 1.0/s;
      s := s + 1.0;
    end;
    {asymptotic expansion}
    if s>1e10 then y := 0.0
    else begin
      z := 1.0/(s*s);
      y := z*PolEvalX(z,BNN,8);
    end;
    y := ln(s) - (0.5/s) - y - w;
  end;
  if subneg then y := y-nz;
  sfc_psi := y;
end;


{---------------------------------------------------------------------------}
function sfc_trigamma(x: extended): extended;
  {-Return the trigamma function psi'(x), INF if x is a negative integer}
var
  s: extended;
begin
  {trigamma = psi(1,x) = zetah(2,x) for x >0}
  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_trigamma := 0.0
    else sfc_trigamma := NaN_x;
  end
  else if x>0.0 then begin
    if x >= 1e20 then sfc_trigamma := 1.0/x
    else sfc_trigamma := sfc_zetah(2.0,x)
  end
  else if frac(x)=0.0 then sfc_trigamma := PosInf_x
  else begin
    {Reflection formula HMF [1] 6.4.7 with n=1}
    {psi(1,1-x) + psi(1,x) = Pi^2*(1+cot(Pi*x)^2) = Pi^2/sin(Pi*x)^2}
    s := sinPi(x);
    sfc_trigamma := PiSqr/sqr(s) - sfc_zetah(2.0,1.0-x);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_tetragamma(x: extended): extended;
  {-Return the tetragamma function psi''(x), NAN/RTE if x is a negative integer}
var
  c,s: extended;
const
  PP3H: THexExtW = ($E0F0,$C4EB,$DAC9,$F80C,$4004);  {2*Pi^3}
begin
  {tetragamma = psi(2,x) = -2*zetah(3,x) for x >0}
  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_tetragamma := 0.0
    else sfc_tetragamma := NaN_x;
  end
  else if x>0.0 then begin
    if x >= 1e20 then begin
      x := 1.0/x;
      sfc_tetragamma := -x*x;
    end
    else sfc_tetragamma := -2.0*sfc_zetah(3.0,x)
  end
  else if frac(x)=0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_tetragamma := NaN_x;
  end
  else begin
    {Reflection formula HMF [1] 6.4.7 with n=2, 2nd derivative of cot(Pi*x) with Maple}
    {psi(2,1-x) - psi(2,x) = 2*Pi^3*cot(Pi*x)*(1+cot(Pi*x)^2)}
    sincosPi(x,s,c);
    c := c/s;
    c := c*(1.0 + c*c);
    s := -2.0*sfc_zetah(3.0,1.0-x);
    sfc_tetragamma := s - extended(PP3H)*c;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_pentagamma(x: extended): extended;
  {-Return the pentagamma function psi'''(x), INF if x is a negative integer}
var
  c,s: extended;
const
  PP4H: THexExtW = ($2C68,$4841,$7461,$C2D1,$4006);  {2*Pi^4}
begin
  {pentagamma = psi(3,x) = 6*zetah(4,x) for x >0}
  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_pentagamma := 0.0
    else sfc_pentagamma := NaN_x;
  end
  else if x>0.0 then begin
    if x >= 1e20 then begin
      x := 1.0/x;
      sfc_pentagamma := 2.0*x*x*x;
    end
    else sfc_pentagamma := 6.0*sfc_zetah(4.0,x);
  end
  else if frac(x)=0.0 then sfc_pentagamma := PosInf_x
  else begin
    {Reflection formula HMF [1] 6.4.7 with n=3, 3rd derivative of cot(Pi*x) with Maple}
    {psi(3,1-x) + psi(3,x) = 2*Pi^4*(1+4*cot(Pi*x)^2 + 3*cot(Pi*x)^4)}
    sincosPi(x,s,c);
    c := sqr(c/s);
    c := 1.0 + c*(4.0 + 3.0*c);
    s := 6.0*sfc_zetah(4.0,1.0-x);
    sfc_pentagamma := extended(PP4H)*c - s;
  end;
end;


{---------------------------------------------------------------------------}
function polygam_negx(n: integer; x: extended): extended;
  {-Polygamma for negative x and 1 <= n <= 12}
var
  c,s,t: extended;
  i,k: integer;
  { for i from 1 to 12 do sort(simplify(expand(diff(cot(x),x$i)))); od;
     -C^2 - 1             with C = cot(x)
     2*C^3 + 2*C
     -6*C^4 -8*C^2 - 2
     24*C^5 + 40*C^3 + 16*C
     -120*C^6 - 240*C^4 - 136*C^2 - 16
     720*C^7 + 1680*C^5 + 1232*C^3 + 272*C
     -5040*C^8 - 13440*C^6 - 12096*C^4 - 3968*C^2 - 272
     40320*C^9 + 120960*C^7 + 129024*C^5 + 56320*C^3 + 7936*C
     -362880*C^10 - 1209600*C^8 - 1491840*C^6 - 814080*C^4 - 176896*C^2 - 7936
     3628800*C^11 + 13305600*C^9 + 18627840*C^7 + 12207360*C^5 + 3610112*C^3 + 353792*C
     -39916800*C^12 - 159667200*C^10 - 250145280*C^8 - 191431680*C^6 - 71867136*C^4 - 11184128*C^2 - 353792
     479001600*C^13 + 2075673600*C^11 + 3597834240*C^9 + 3149752320*C^7 + 1436058624*C^5 + 309836800*C^3 + 22368256*C
  }
const
  ca: array[0..53] of single = ({Above coefficients reflected and positive,}
        { 1} 1,1,               {all are exactly representable as single!}
        { 2} 2,2,
        { 3} 2,8,6,
        { 4} 16,40,24,
        { 5} 16,136,240,120,
        { 6} 272,1232,1680,720,
        { 7} 272,3968,12096,13440,5040,
        { 8} 7936,56320,129024,120960,40320,
        { 9} 7936,176896,814080,1491840,1209600,362880,
        {10} 353792,3610112,12207360,18627840,13305600,3628800,
        {11} 353792,11184128,71867136,191431680,250145280,159667200,39916800,
        {12} 22368256,309836800,1436058624,3149752320.0,3597834240.0,2075673600,479001600);
  ko: array[1..12] of integer = (0,2,4,7,10,14,18,23,28,34,40,47); {index of first coeff for n}
begin
  sincosPi(x,s,c);
  if (n<1) or (n>12) or (x>=0.0) or (s=0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    polygam_negx := NaN_x;
    exit;
  end;
  {Reflection formula: HMF[1] (6.4.7) or  NIST[30], (5.15.6)}
  {psi(n,1-x) + (-1)^(n+1)*psi(n,x) = (-1)^n*Pi*d^n/dx^n(cot(Pi*x))}

  {psi(n,x) = - (-1)^(n+1)*psi(n,1-x) + (-1)^(n+n+1)*Pi*d^n/dx^n(cot(Pi*x))}
  {psi(n,x) = (-1)^n*psi(n,1-x) - Pi*d^n/dx^n(cot(Pi*x))}
  c := c/s;
  s := c*c;
  k := ko[n];
  t := 0.0;
  {Compute polynomial in cot(Pi*x)^2}
  for i := (succ(n) div 2) downto 0 do t := t*s + ca[k+i];
  s := sfc_polygamma(n,1.0-x);
  if odd(n) then s := -s
  else begin
    {t is already correct for odd n, for even n multiply with -cot(Pi*x), the}
    {minus compensates the fact that the coefficients in ca are all positive.}
    t := -t*c;
  end;
  t := t*intpower(Pi,n+1);
  polygam_negx := s + t;
end;


{---------------------------------------------------------------------------}
function polygam_sum(n: integer; x: extended): extended;
  {-Direct summation polygamma(n,x) = n!/x^(n+1)*(1 + sum(1/(1+k/x)^(n+1), k=1...))}
const
  KMAX = 200;
var
  s,t,e: extended;
  k,m: integer;
begin
  m := -(n+1);
  e := 0.125*eps_x;
  s := 0.0;
  k := 1;
  repeat
    t := power(1.0+k/x, m);
    s := s + t;
    inc(k);
  until (t<=e) or (k>KMAX);
  if k>KMAX then begin
    {no convergence}
    polygam_sum := NaN_x;
  end
  else begin
    t := sfc_lngamma(n+1) - (n+1)*ln(x);
    if t+ln1p(s) >= ln_MaxExt then polygam_sum := PosInf_x
    else polygam_sum := (1.0+s)*exp(t);
  end;
end;


{---------------------------------------------------------------------------}
function polygam_ase(n: integer; x: extended): extended;
  {-Polygamma asymptotic expansion}
var
  z,f,t,b: extended;
  k,j: integer;
const
  K2MAX =  120; {doubled maximum index in asymptotic sum}
  NPMAX = 2000; {Max. n for calculation of (n-1)!/x^n as product}
  TFMAX = 2.6E4913;  {~ 2*MaxExtended*eps_s, used for 'no-convergence' check}
begin
  f := sfc_lngamma(n);
  z := n*ln(x);
  t := f-z;
  if t > ln_MaxExt then z := PosInf_x
  else if t < ln_MinExt then z := 0.0
  else begin
    {calculate f = (n-1)!/x^n}
    if (abs(z)<11356.0) and (n<MAXGAMX-1) then begin
      f := sfc_fac(n-1)*power(x,-n);
    end
    else begin
      if n <= NPMAX then begin
        {avoid exp/ln inaccuracies, example: sfc_polygamma(1800,980)}
        f := 1.0;
        for k:=1 to n-1 do f := f*k/x;
        f := f/x;
      end
      else f := exp(t);
    end;
    {First two terms}
    z := f + 0.5*f*(n/x);
    {Add sum terms with Bernoulli numbers}
    k := 2;
    repeat
      j := k-1;
      t := (n+j)/k*(n+j-1)/j/x/x;
      b := sfc_bernoulli(k);
      if (k>K2MAX) or (abs(f) > abs(TFMAX/t/b)) then begin
        {Convergence failed if k too large or a summand ~> MaxExtended}
        {Try direct summation, complete failure if result is NaN}
        z := polygam_sum(n,x);
        {$ifopt R+}
          if (IsNaN(z)) and (RTE_NoConvergence>0) then RunError(byte(RTE_NoConvergence));
        {$endif}
        polygam_ase := z;
        exit;
      end;
      f := f*t;
      t := f*b;
      z := z + t;
      k := k+2;
    until abs(t)<abs(z)*eps_x;
  end;
  polygam_ase := z
end;


{---------------------------------------------------------------------------}
function sfc_polygamma(n: integer; x: extended): extended;
  {-Return the polygamma function: n'th derivative of psi; n>=0, x>0 for n>12.}
  { Note: Accuracy may be reduced for n>=MAXGAMX due to ln/exp operations.}
var
  z,f,t: extended;
  k: integer;
const
  NMAX  =   12; {Max. n for negative x}
  NPMAX = 2500; {Max. n > 1752 for calculation of n! as product}
begin
  if n<4 then begin
    case n of
        0: sfc_polygamma := sfc_psi(x);
        1: sfc_polygamma := sfc_trigamma(x);
        2: sfc_polygamma := sfc_tetragamma(x);
        3: sfc_polygamma := sfc_pentagamma(x);
      else begin
             {RTE or NAN if n < 0}
             {$ifopt R+}
               if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
             {$endif}
             sfc_polygamma := NaN_x;
           end;
    end;
  end
  else if IsNanOrInf(x) then begin
    if x=PosInf_x then begin
      if n>0 then sfc_polygamma := 0.0
      else sfc_polygamma := PosInf_x
    end
    else sfc_polygamma := NaN_x;
  end
  else if x > 0.0 then begin
    {polygamma(n,x) = (-1)^(n+1)*n!*sum(1/(x+k)^(n+1), k>=0), HMF[1] 6.4.10}
    {or with the Hurwitz zeta:  polygamma(n,x) = (-1)^(n+1)*n!*zetah(n+1,x)}
    z := sfc_zetah(n+1,x);
    if z=0.0 then begin
      {If zetah=0, use HMF[1] 6.4.10 or the asymptotic formula 6.4.11}
      if x>0.25*n then z := polygam_ase(n,x)
      else z := polygam_sum(n,x)
    end
    else if n<MAXGAMX-1 then begin
      {n! is OK, but n!*zetah may overflow for small x}
      f := sfc_fac(n);
      t := ln(f)+ln(z);
      if t<ln_MaxExt then z := z*f
      else z := PosInf_x;
    end
    else begin
      f := sfc_lngamma(n+1);
      t := f + ln(z);
      if t > ln_MaxExt then z := PosInf_x
      else if t < ln_MinExt then z := 0.0
      else begin
        if n>NPMAX then z := exp(t)
        else begin
          {example: sfc_polygamma(2001,32)}
          z := sfc_fac(1752)*z;
          for k:=1753 to n do z := z*k;
        end;
      end;
    end;
    if odd(n) then sfc_polygamma := z
    else sfc_polygamma := -z;
  end
  else if (x<0.0) and (n>nmax) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_polygamma := NaN_x;
  end
  else begin
    {x<0 and n<=nmax}
    sfc_polygamma := polygam_negx(n,x);
  end;
end;

{------------------------------------------}
{------------  Beta functions -------------}
{------------------------------------------}

{---------------------------------------------------------------------------}
function sfc_lnbeta(x,y: extended): extended;
  {-Return the logarithm of |beta(x,y)|=|gamma(x)*gamma(y)/gamma(x+y)|}
var
  c,s: extended;
begin
  if IsNanOrInf(x) or IsNanOrInf(x) then begin
    sfc_lnbeta := NaN_x;
    exit;
  end;
  {based on SLATEC [14] routine dlbeta.f}
  {force x <= y}
  if x>y then begin
    c := x;
    x := y;
    y := c;
  end;
  s := x+y;
  if x >= XLGAE then begin
    {x and y >= XLGAE}
    c := sfc_lngcorr(x) + sfc_lngcorr(y) - sfc_lngcorr(s);
    sfc_lnbeta := -0.5*ln(y) + lnsqrt2pi + c + (x-0.5)*ln(x/s) + y*ln1p(-x/s);
  end
  else if (y >= XLGAE) and (s >= XLGAE) then begin
    {Fixed in 1.02.01: check s >= XLGAE, otherwise lnbeta(-10.1,10) fails}
    {x < XLGAE, y >= XLGAE}
    c := sfc_lngcorr(y) - sfc_lngcorr(s);
    sfc_lnbeta := sfc_lngamma(x) + c + x - x*ln(s) + (y-0.5)*ln1p(-x/s)
  end
  else begin
    {x,y < XLGAE; or s<=0}
    sfc_lnbeta := sfc_lngamma(x) + sfc_lngamma(y) - sfc_lngamma(s);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_beta(x,y: extended): extended;
  {-Return the function beta(x,y)=gamma(x)*gamma(y)/gamma(x+y)}
var
  r,t: extended;
begin
  {If one arg is 1, return the inverse of the other because}
  {gamma(1)*gamma(y)/gamma(y+1) = 1/y, analogue for y=1}
  if IsNanOrInf(x) or IsNanOrInf(x) then sfc_beta := NaN_x
  else if x=1.0 then sfc_beta := 1.0/y
  else if y=1.0 then sfc_beta := 1.0/x
  else begin
    {force x >= y}
    if x<y then begin
      t := x;
      x := y;
      y := t;
    end;
    t := x+y;
    if (t<=0.0) and (frac(t)=0.0) then begin
      if (frac(x)<>0.0) and (frac(y)<>0.0) then begin
        {gamma(x+y) = Inf, gamma(x)*gamma(y)<>Inf}
        sfc_beta := 0.0;
      end
      else begin
        t := sfc_gamma(x);
        r := sfc_pochhammer(y,x);
        sfc_beta := t/r;
      end;
      exit;
    end;
    r := sfc_lnbeta(x,y);
    t := exp(r);
    if (x<0.0) or (y<0.0) then begin
      {get signs only if necessary}
      r := sfc_signgamma(x)*sfc_signgamma(y)/sfc_signgamma(x+y);
      sfc_beta := r*t;
    end
    else sfc_beta := t
  end;
end;

{$ifdef VER50}
{---------------------------------------------------------------------------}
function sfc_ibetaprefix(a,b,x,y: extended): extended;
  {-Return (x^a)(y^b)/Beta(a,b)}
var
  w,z,t: extended;
begin
  w := a*ln(x);
  if abs(x) < 0.5 then z := b*ln1p(-x)
  else z := b*ln(y);
  t := sfc_lnbeta(a,b);
  t := w+z-t;
  sfc_ibetaprefix:= exp(t);
end;

{$else}

{---------------------------------------------------------------------------}
function sfc_ibetaprefix(a,b,x,y: extended): extended;
  {-Return (x^a)(y^b)/Beta(a,b) using Lanczos approximation}
var
  res,c,
  agh,bgh,cgh,b1,b2,
  l,l1,l2,l3,ratio: extended;
  sa: boolean;
begin

  {Ref: Boost [19], beta.hpp, function ibeta_power_terms}
  c := a + b;
  {combine power terms with Lanczos approximation}
  agh := a + lanczos_gm05;
  bgh := b + lanczos_gm05;
  cgh := c + lanczos_gm05;
  res := lanczos(c,true)/(lanczos(a,true)*lanczos(b,true));

  {l1 and l2 are the base of the exponents minus one}
  l1 := (x*b - y*agh) / agh;
  l2 := (y*a - x*bgh) / bgh;
  if minx(abs(l1), abs(l2)) < 0.2 then begin
    {when the base of the exponent is very near 1 we get really}
    {gross errors unless extra care is taken}
    if (l1*l2 > 0) or (minx(a, b) < 1) then begin
      if abs(l1) < 0.1 then res := res*exp(a*ln1p(l1))
      else res := res*power((x * cgh) / agh, a);
      if abs(l2) < 0.1 then res := res*exp(b*ln1p(l2))
      else res := res*power((y * cgh) / bgh, b);
    end
    else if maxx(abs(l1), abs(l2)) < 0.5 then begin
      sa := a < b;
      ratio := b / a;
      if (sa and (ratio*l2 < 0.1)) or ((not sa) and (l1/ratio > 0.1)) then begin
        l3 := expm1(ratio*ln1p(l2));
        l3 := l1 + l3 + l3*l1;
        l3 := a*ln1p(l3);
        res := res*exp(l3);
      end
      else begin
        l3 := expm1(ln1p(l1) / ratio);
        l3 := l2 + l3 + l3*l2;
        l3 := b*ln1p(l3);
        res := res*exp(l3);
      end;
    end
    else if abs(l1) < abs(l2) then begin
      {First base near 1 only}
      l := a*ln1p(l1) + b*ln((y*cgh) / bgh);
      res := res*exp(l);
    end
    else begin
      {Second base near 1 only}
      l := b*ln1p(l2) + a*ln((x*cgh) / agh);
      res := res*exp(l);
    end
  end
  else begin
    {general case:}
    b1 := (x*cgh) / agh;
    b2 := (y*cgh) / bgh;
    l1 := a*ln(b1);
    l2 := b*ln(b2);
    if (l1 >= ln_MaxExt) or (l1 <= ln_MinExt) or (l2 >= ln_MaxExt) or (l2 <= ln_MinExt) then begin
      {Oops, overflow, sidestep}
      if a < b then res := res*power(power(b2, b/a)*b1, a)
      else res := res*power(power(b1, a/b)*b2, b);
    end
    else begin
      {finally the normal case}
      res := res*power(b1,a)*power(b2,b);
    end;
  end;
  {combine with the leftover terms from the Lanczos approximation}
  res := res*sqrt(bgh/exp(1));
  res := res*sqrt(agh/cgh);
  sfc_ibetaprefix:= res;
end;
{$endif}


{---------------------------------------------------------------------------}
function ibeta_series(a, b, x: extended; normalised: boolean): extended;
  {-Power series for incomplete beta integral. Use when b*x is small }
var
  s, t, u, v, t1, z, ai: extended;
  n: integer;
const
  nmax = 30000;
begin
  {Ref: Cephes [7], function pseriesl in ldouble/incbetl.c}
  ai := 1.0/a;
  u  := (1.0-b)*x;
  v  := u/(a+1.0);
  t1 := v;
  t  := u;
  n  := 2;
  s  := 0.0;
  z  := abs(t1+ai)*eps_x;
  if z=0.0 then z := sqr(eps_x);
  repeat
    u := (n - b)*x/n;
    t := t*u;
    v := t/(a + n);
    s := s + v;
    n := n + 1;
  until (abs(v) <= z) or (n>nmax);

  if (n>=nmax) and (RTE_NoConvergence>0) then begin
    RunError(byte(RTE_NoConvergence));
  end;

  s := (s + t1) + ai;
  if normalised then begin
    if a+b < MAXGAMX then begin
      t := sfc_beta(a,b);
      ibeta_series := s/t*power(x, a);
    end
    else begin
      u := a*ln(x);
      s := ln(s);
      t := sfc_lnbeta(a,b);
      ibeta_series := exp((u+s)-t);
    end;
  end
  else begin
    {no overflow because x <= 1, a > 0}
    ibeta_series := s*power(x, a);
  end;
end;


{---------------------------------------------------------------------------}
function ibeta_cf(a, b, x: extended; type2: boolean): extended;
  {-Continued fraction expansion type #1 or #2 for incomplete beta integral}
var
  xk, pk, pkm1, pkm2, qk, qkm1, qkm2: extended;
  k1, k2, k3, k4, k5, k6, k7, k8: extended;
  dk, r, t, ans, big, thresh: extended;
  i,imax: integer;
begin
  {Ref: Cephes [7], functions incbcfl & incbdl in ldouble/incbetl.c}
  {The two continued fraction from [7] are almost identical and are}
  {merged into a single routine, selected by the type2 parameter.}
  k1 := a;
  k3 := a;
  k4 := a + 1.0;
  k5 := 1.0;
  k7 := k4;
  k8 := a + 2.0;

  pkm2 := 0.0;
  qkm2 := 1.0;
  pkm1 := 1.0;
  qkm1 := 1.0;

  ans  := 1.0;
  big  := ldexp(1,63);

  r := 1.0;
  t := sqrt(maxx(a,b));
  if t<400.0 then t := 400.0
  else if t>32000.0 then t := 32000.0;
  imax := round(t);

  thresh := 3.0*eps_x;
  if type2 then begin
    k2 := b - 1.0;
    k6 := a + b;
    dk := -1.0;
    x  := x/(1.0 - x);
  end
  else begin
    k2 := a + b;
    k6 := b - 1.0;
    dk := 1.0;
  end;

  for i:=1 to imax do begin
    xk := - (x * k1 * k2) / (k3 * k4);
    pk := pkm1 + pkm2 * xk;
    qk := qkm1 + qkm2 * xk;
    pkm2 := pkm1;
    pkm1 := pk;
    qkm2 := qkm1;
    qkm1 := qk;

    xk := (x * k5 * k6) / (k7 * k8);
    pk := pkm1 + pkm2 * xk;
    qk := qkm1 + qkm2 * xk;
    pkm2 := pkm1;
    pkm1 := pk;
    qkm2 := qkm1;
    qkm1 := qk;

    if qk <> 0.0 then r := pk / qk;
    if r <> 0.0 then begin
      t := abs((ans - r) / r);
      ans := r;
    end
    else t := 1.0;

    if t < thresh then begin
      ibeta_cf := ans;
      exit;
    end;

    k1 := k1 + 1.0;
    k2 := k2 + dk;
    k3 := k3 + 2.0;
    k4 := k4 + 2.0;
    k5 := k5 + 1.0;
    k6 := k6 - dk;
    k7 := k7 + 2.0;
    k8 := k8 + 2.0;

    if abs(qk) + abs(pk) > big then begin
      pkm2 := pkm2 * eps_x;
      pkm1 := pkm1 * eps_x;
      qkm2 := qkm2 * eps_x;
      qkm1 := qkm1 * eps_x;
    end;
    if (abs(qk) < eps_x) or (abs(pk) < eps_x) then begin
      pkm2 := pkm2 * big;
      pkm1 := pkm1 * big;
      qkm2 := qkm2 * big;
      qkm1 := qkm1 * big;
    end;

  end;

  if RTE_NoConvergence>0 then begin
    RunError(byte(RTE_NoConvergence));
  end;
  ibeta_cf := ans;
end;


{---------------------------------------------------------------------------}
function basym(a,b,lambda,eps: extended): extended;
  {-Return I_x(a,b for large a,b. lambda = (a+b)*(1-x) - b, eps is}
  { the tolerance used. It is assumed that lambda is nonnegative }
  { and that a and b are greater than or equal to 15.}
  { -1 is returned if no convergence for array size num}
const
  num = 20; {maximum n value in the main repeat loop, must be even.}
  e0  = 1.128379167095512573896; {2/sqrt(Pi)}
  e1  = 0.353553390593273762200; {2^(-3/2)}
var
  a0,b0,c,d: array[0..num] of extended;
var
  h,r,s,t,w,j0,j1,h2,r0,r1,t0,t1,w0,z0,z2,hn,zn,sum,znm1,bsum,dsum: extended;
  i,j,m,n,im1,mm1,mmj: integer;
  done: boolean;
begin
  {Ref: [37] Algorithm 708, function BASYM}
  {ln1pmx = -rlog1}
  h := -a*ln1pmx(-lambda/a);
  t := -b*ln1pmx(lambda/b);
  h := h+t;
  t := exp(-h);
  if t=0.0 then begin
    basym := 0.0;
    exit;
  end;

  z0  := sqrt(h);
  z2  := 2.0*h;
  znm1:= 0.5*z0/e1;
  if a<b then begin
    h  := a/b;
    r0 := 1.0/(1.0+h);
    r1 := (b-a)/b;
    w0 := 1.0/sqrt(a*(1.0+h));
  end
  else begin
    h  := b/a;
    r0 := 1.0/(1.0+h);
    r1 := (b-a) / a;
    w0 := 1.0/sqrt(b*(1.0+h));
  end;

  a0[0]:= (2.0*r1)/THREE;
  c[0] := -0.5*a0[0];
  d[0] := -c[0];
  j0   := sfc_erfce(z0)*0.5/e0;
  j1   := e1;
  sum  := j0 + d[0]*w0*j1;
  s    := 1.0;
  h2   := h*h;
  hn   := 1.0;
  w    := w0;
  zn   := z2;

  n := 2;
  repeat
    hn := h2*hn;
    s  := s + hn;
    a0[n-1]:= 2.0*r0*(1.0+h*hn)/(n+2);
    a0[n]  := 2.0*r1*s/(n+3);

    for i:=n to n+1 do begin
      im1 := pred(i);
      r   := -0.5*succ(i);
      b0[0] := r*a0[0];
      for m:=2 to i do begin
        bsum := 0.0;
        mm1  := m-1;
        for j:=1 to mm1 do begin
          mmj := m-j;
          bsum := bsum + (j*r - mmj)*a0[j-1]*b0[mmj-1];
        end;
        b0[mm1] := r*a0[mm1] + bsum/m;
      end;
      c[im1] := b0[im1]/succ(i);
      dsum := 0.0;
      for j:=1 to im1 do dsum := dsum + d[im1-j]*c[j-1];
      d[im1] := -(dsum + c[im1]);
    end;

    j0 := e1*znm1 + (n-1)*j0;
    j1 := e1*zn   + n*j1;
    znm1 := z2*znm1;
    zn := z2*zn;
    w  := w0*w;
    t0 := d[n-1]*w*j0;
    w  := w0*w;
    t1 := d[n]*w*j1;
    sum := sum + (t0+t1);
    inc(n,2);
    done := abs(t0) + abs(t1) <= eps*sum;
  until done or (n>num);

  if done then begin
    {compute w = exp(-bcorr(a,b)) using sfc_lngcorr}
    w := sfc_lngcorr(b) - sfc_lngcorr(a+b);
    w := w + sfc_lngcorr(a);
    w := exp(-w);
    basym := e0*t*w*sum;
  end
  else basym := -1;
end;


{---------------------------------------------------------------------------}
function ccdf_binom(n,k: integer; x,y: extended): extended;
  {-Return complement of the binomial distribution}
var
  s,t,p: extended;
  i: integer;
begin
  {Ref: Boost [19], beta.hpp, function binomial_ccdf}
  p := power(x,n);
  s := 1.0;
  if p>0.0 then begin
    t := s;
    for i:=pred(n) downto succ(k) do begin
      t := t*(succ(i)*y);
      t := t/((n-i)*x);
      s := s + t;
    end;
  end;
  ccdf_binom := s*p;
end;


{---------------------------------------------------------------------------}
function sfc_ibeta(a, b, x: extended): extended;
  {-Return the normalised incomplete beta function, a>0, b>0, 0 <= x <= 1}
  { sfc_ibeta = integral(t^(a-1)*(1-t)^(b-1) / beta(a,b), t=0..x)}
var
  a1, b1, x1, t, w, xc, y: extended;
  flag : boolean;
begin

  {Refs:  Cephes[7], function incbetl in ldouble/incbetl.c}
  {and  Boost [19], beta.hpp; [37] Alg.708, function BASYM}

  if (a <= 0.0) or (b <= 0.0) or (x < 0.0) or (x > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_ibeta := NaN_x;
    exit;
  end;

  if (x=0.0) or (x=1.0) then begin
    sfc_ibeta := x;
    exit;
  end
  else if b=1.0 then begin
    sfc_ibeta := power(x,a);
    exit;
  end
  else if (a<1e-22) and (b<1e-22) then begin
   sfc_ibeta  := b/(a+b);
   exit;
  end;

  flag := false;
  w := 1.0 - x;

  {here t corresponds to lambda from [19] and [37]}
  if a>b then t := a - (a+b)*x
  else t := (a+b)*w - b;

  {reverse a and b if x is greater than the mean}
  if t < 0 then begin
    flag := true;
    a1 := b;
    b1 := a;
    xc := x;
    x1 := w;
    t  := -t;
  end
  else begin
    a1 := a;
    b1 := b;
    xc := w;
    x1 := x;
  end;

  if (a>=100.0) and (b>=100.0) then begin
    {Check if asymptotic expansion can be used}
    if ((a1<=b1) and (t<=0.03*a1)) or ((a1>b1) and (t<=0.03*b1)) then begin
      t := basym(a1,b1,t,100.0*eps_x);
      if t>=0.0 then begin
        {convergence with current basym array size}
        if flag then t := 1.0 - t;
        sfc_ibeta := t;
        exit;
      end;
    end;
  end;

  if (b1<=40) and (frac(b1)=0) and (a1<=32000) and (frac(a1)=0) then begin
    {Use finite sum related to binomial distribution}
    t := ccdf_binom(round(a1+b1-1),round(a1-1), x1, xc);
  end
  else if (b1*x1 <= 1.0) and (x1 <= 0.95) then begin
    t := ibeta_series(a1, b1, x1, true);
  end
  else begin
    {choose expansion for optimal convergence}
    y := x1*(a1 + b1 - 2.0) - (a1 - 1.0);
    if y < 0.0 then w := ibeta_cf(a1, b1, x1, false)
    else w := ibeta_cf(a1, b1, x1, true) / xc;
    {$ifdef VER50}
       {multiply w by the factor x^a*(1-x)^b*gamma(a+b)/(a*gamma(a)*gamma(b))}
       if flag then begin
         {here xc=x and x1=1-x}
         y := a1*ln1p(-x);
         t := b1*ln(x);
       end
       else begin
         {here x1=x and xc=1-x}
         y := a1*ln(x);
         t := b1*ln1p(-x);
       end;
       if (a1+b1 < MAXGAMX) and (abs(y) < ln_MaxExt) and (abs(t) < ln_MaxExt) then begin
         t := power(xc, b1) ;
         y := power(x1, a1);
         t := (t*y/a1)*w;
         t := t/sfc_beta(a1,b1);
       end
       else begin
         {resort to logarithms}
         t := t - sfc_lnbeta(a1,b1);
         y := y + ln(w/a1);
         t := exp(y+t);
       end;
    {$else}
      t := sfc_ibetaprefix(a1,b1,x1,xc);
      t := w*(t/a1);
    {$endif}
  end;

  if flag then sfc_ibeta := 1.0 - t
  else sfc_ibeta := t;

end;


{---------------------------------------------------------------------------}
function sfc_nnbeta(a, b, x: extended): extended;
  {-Return the non-normalised incomplete beta function B_x(a,b)}
  { for 0<=x<=1, B_x = integral(t^(a-1)*(1-t)^(b-1), t=0..x).  }
var
  a1, b1, x1, t, w, xc, y: extended;
  flag : boolean;
begin
  {Refs:  Cephes[7], function incbetl in ldouble/incbetl.c}
  {and  Boost [19], beta.hpp; [37] Alg.708, function BASYM}
  if (x < 0.0) or (x > 1.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_nnbeta := NaN_x;
    exit;
  end;

  if x=0.0 then begin
    sfc_nnbeta := 0.0;
    exit;
  end
  else if x=1.0 then begin
    sfc_nnbeta := sfc_beta(a,b);
    exit;
  end
  else if a=1.0 then begin
    sfc_nnbeta := -pow1pm1(-x,b)/b;
    exit;
  end
  else if b=1.0 then begin
    sfc_nnbeta := power(x,a)/a;
    exit;
  end;

  if (a<=0.0) or (b<=0.0) then begin
    if (a>0.0) or (frac(a)<>0.0) then begin
      {http://functions.wolfram.com/06.19.26.0005.01}
      y := sfc_2F1(a,1.0-b,a+1.0,x);
      w := power(x,a);
      sfc_nnbeta := (w/a)*y;
    end
    else begin
      {here: a is a negative integer or zero}
      if (b>0.0) or (frac(b)<>0.0) then begin
        {http://functions.wolfram.com/06.19.26.0007.01}
        y := sfc_2F1(1.0,a+b,b+1.0,1.0-x);
        w := power(1.0-x, b);
        t := power(x,a);
        y := (w*t/b)*y;
        t := sfc_beta(a,b);
        sfc_nnbeta := t-y;
      end
      else begin
        {$ifopt R+}
          if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
        {$endif}
        sfc_nnbeta := NaN_x;
      end;
    end;
    exit;
  end;

  flag := false;
  w := 1.0 - x;

  {here t corresponds to lambda from [19] and [37]}
  if a>b then t := a - (a+b)*x
  else t := (a+b)*w - b;

  {reverse a and b if x is greater than the mean}
  if t < 0 then begin
    flag := true;
    a1 := b;
    b1 := a;
    xc := x;
    x1 := w;
    t  := -t;
  end
  else begin
    a1 := a;
    b1 := b;
    xc := w;
    x1 := x;
  end;

  if (a>=100.0) and (b>=100.0) then begin
    {Check if asymptotic expansion can be used}
    if ((a1<=b1) and (t<=0.03*a1)) or ((a1>b1) and (t<=0.03*b1)) then begin
      t := basym(a1,b1,t,100.0*eps_x);
      if t>=0.0 then begin
        {convergence with current basym array size}
        if flag then t := 1.0 - t;
        w := sfc_beta(a1,b1);
        sfc_nnbeta := w*t;
        exit;
      end;
    end;
  end;

  if (b1*x1 <= 1.0) and (x1 <= 0.95) then begin
    t := ibeta_series(a1, b1, x1, false);
  end
  else begin
    {choose expansion for optimal convergence}
    y := x1*(a1 + b1 - 2.0) - (a1 - 1.0);
    if y < 0.0 then begin
      w := ibeta_cf(a1, b1, x1, false);
      t := power(xc, b1);
    end
    else begin
      w := ibeta_cf(a1, b1, x1, true);
      t := power(xc, b1 - 1.0);
    end;
    y := power(x1, a1);
    t := (t*y/a1)*w;
  end;

  if flag then begin
    {Anti Ver50 code}
    w := sfc_beta(a1,b1);
    sfc_nnbeta := w - t;
  end
  else sfc_nnbeta := t;

end;


{---------------------------------------------------------------------------}
{------------------- Incomplete Gamma function -----------------------------}
{---------------------------------------------------------------------------}

{---------------------------------------------------------------------------}
procedure igam_pqasymp(a,x: extended; var p,q: extended);
  {-Incomplete gamma functions for large a and a ~ x}
var
  y,z: extended;
  w: array[0..12] of extended;
const
  N0=18;
  C0: array[0..N0] of extended = (
       -0.333333333333333333333,
        0.0833333333333333333333,
       -0.0148148148148148148148,
        0.00115740740740740740741,
        0.000352733686067019400353,
       -0.0001787551440329218107,
        0.39192631785224377817e-4,
       -0.218544851067999216147e-5,
       -0.18540622107151599607e-5,
        0.829671134095308600502e-6,
       -0.176659527368260793044e-6,
        0.670785354340149858037e-8,
        0.102618097842403080426e-7,
       -0.438203601845335318655e-8,
        0.914769958223679023418e-9,
       -0.255141939949462497669e-10,
       -0.583077213255042506746e-10,
        0.243619480206674162437e-10,
       -0.502766928011417558909e-11);

  N1=16;
  C1: array[0..N1] of extended = (
       -0.00185185185185185185185,
       -0.00347222222222222222222,
        0.00264550264550264550265,
       -0.000990226337448559670782,
        0.000205761316872427983539,
       -0.40187757201646090535e-6,
       -0.18098550334489977837e-4,
        0.764916091608111008464e-5,
       -0.161209008945634460038e-5,
        0.464712780280743434226e-8,
        0.137863344691572095931e-6,
       -0.575254560351770496402e-7,
        0.119516285997781473243e-7,
       -0.175432417197476476238e-10,
       -0.100915437106004126275e-8,
        0.416279299184258263623e-9,
       -0.856390702649298063807e-10);

   N2=14;
   C2: array[0..N2] of extended = (
         0.00413359788359788359788,
        -0.00268132716049382716049,
         0.000771604938271604938272,
         0.200938786008230452675e-5,
        -0.000107366532263651605215,
         0.529234488291201254164e-4,
        -0.127606351886187277134e-4,
         0.342357873409613807419e-7,
         0.137219573090629332056e-5,
        -0.629899213838005502291e-6,
         0.142806142060642417916e-6,
        -0.204770984219908660149e-9,
        -0.140925299108675210533e-7,
         0.622897408492202203356e-8,
        -0.136704883966171134993e-8);

   N3=12;
   C3: array[0..N3] of extended = (
         0.000649434156378600823045,
         0.000229472093621399176955,
        -0.000469189494395255712128,
         0.000267720632062838852962,
        -0.756180167188397641073e-4,
        -0.239650511386729665193e-6,
         0.110826541153473023615e-4,
        -0.56749528269915965675e-5,
         0.142309007324358839146e-5,
        -0.278610802915281422406e-10,
        -0.169584040919302772899e-6,
         0.809946490538808236335e-7,
        -0.191111684859736540607e-7);

   N4=10;
   C4: array[0..N4] of extended = (
        -0.000861888290916711698605,
         0.000784039221720066627474,
        -0.000299072480303190179733,
        -0.146384525788434181781e-5,
         0.664149821546512218666e-4,
        -0.396836504717943466443e-4,
         0.113757269706784190981e-4,
         0.250749722623753280165e-9,
        -0.169541495365583060147e-5,
         0.890750753220530968883e-6,
        -0.229293483400080487057e-6);

   N5=8;
   C5: array[0..N5] of extended = (
        -0.000336798553366358150309,
        -0.697281375836585777429e-4,
         0.000277275324495939207873,
        -0.000199325705161888477003,
         0.679778047793720783882e-4,
         0.141906292064396701483e-6,
        -0.135940481897686932785e-4,
         0.801847025633420153972e-5,
        -0.229148117650809517038e-5);

   N6=10;
   C6: array[0..N6] of extended = (
         0.000531307936463992223166,
        -0.000592166437353693882865,
         0.000270878209671804482771,
         0.790235323266032787212e-6,
        -0.815396936756196875093e-4,
         0.561168275310624965004e-4,
        -0.183291165828433755673e-4,
        -0.307961345060330478256e-8,
         0.346515536880360908674e-5,
        -0.20291327396058603727e-5,
         0.57887928631490037089e-6);

   N7=8;
   C7: array[0..N7] of extended = (
         0.000344367606892377671254,
         0.517179090826059219337e-4,
        -0.000334931610811422363117,
         0.000281269515476323702274,
        -0.000109765822446847310235,
        -0.127410090954844853795e-6,
         0.277444515115636441571e-4,
        -0.182634888057113326614e-4,
         0.578769494973505239894e-5);

   N8=6;
   C8: array[0..N8] of extended = (
        -0.000652623918595309418922,
         0.000839498720672087279993,
        -0.000438297098541721005061,
        -0.696909145842055197137e-6,
         0.000166448466420675478374,
        -0.000127835176797692185853,
         0.462995326369130429061e-4);

   N9=4;
   C9: array[0..N9] of extended = (
        -0.000596761290192746250124,
        -0.720489541602001055909e-4,
         0.000678230883766732836162,
        -0.0006401475260262758451,
         0.000277501076343287044992);

   N10=2;
   C10: array[0..N10] of extended = (
          0.00133244544948006563713,
         -0.0019144384985654775265,
          0.00110893691345966373396);

   N11=4;
   C11: array[0..N11] of extended = (
          0.00157972766073083495909,
          0.000162516262783915816899,
         -0.00206334210355432762645,
          0.00213896861856890981541,
         -0.00101085593912630031708);

   N12=2;
   C12: array[0..2] of extended = (
         -0.00407251211951401664727,
          0.00640336283380806979482,
         -0.00404101610816766177474);
begin
   {Asymptotic expansions of incomplete gamma functions P(a,x) and Q(a,x)}
   {when a is large and x ~ a, }

   {Based on \boost\math\special_functions\detail\igamma_large.hpp [19]}
   {Copyright John Maddock 2006, see 3rdparty.ama for Boost license}

   p := -ln1pmx((x-a)/a);
   y := a*p;
   z := sqrt(2.0*p);
   if x<a then z := -z;

   w[0] := PolEvalX(z, C0, 1+N0 );
   w[1] := PolEvalX(z, C1, 1+N1 );
   w[2] := PolEvalX(z, C2, 1+N2 );
   w[3] := PolEvalX(z, C3, 1+N3 );
   w[4] := PolEvalX(z, C4, 1+N4 );
   w[5] := PolEvalX(z, C5, 1+N5 );
   w[6] := PolEvalX(z, C6, 1+N6 );
   w[7] := PolEvalX(z, C7, 1+N7 );
   w[8] := PolEvalX(z, C8, 1+N8 );
   w[9] := PolEvalX(z, C9, 1+N9 );
   w[10]:= PolEvalX(z, C10,1+N10);
   w[11]:= PolEvalX(z, C11,1+N11);
   w[12]:= PolEvalX(z, C12,1+N12);

   p := PolEvalX(1.0/a, w, 13);
   p := p*exp(-y)/sqrt(a*TwoPi);
   if x<a then p := -p;
   p := p + 0.5*sfc_erfc(sqrt(y));
   if x>=a then begin
     q := p;
     p := 1.0 - q;
   end
   else begin
     q := 1.0 - p;
   end;
end;


{---------------------------------------------------------------------------}
function igam_aux(x: extended): extended;
  {-Temme's function g(x) in 1/Gamma(x) = 1 + x*(x-1)*g(x), x >= -1}
const
  {Maple definition: g := x -> (1/GAMMA(x+1)-1)/(x*(x-1));}
  {chebyshev(g(x), x=0..1, 0.5e-20); calculated with Digits:=30;}
  gan = 17;
  gah : array[0..gan-1] of THexExtW = (
          ($239C,$09A0,$F2BB,$81BD,$BFFF),  {-1.0136092580098657770}
          ($CD4B,$0359,$8CDD,$A0BF,$3FFB),  {+7.8490353102478228353E-2}
          ($D4FE,$74A4,$7C17,$DD79,$3FF7),  {+6.7588668743258315530E-3}
          ($1E68,$B149,$93E4,$A7A5,$BFF5),  {-1.2790434869623468121E-3}
          ($500F,$828F,$C963,$C22B,$3FF0),  {+4.6293983864273958470E-5}
          ($95EA,$C104,$94B3,$9190,$3FED),  {+4.3381681744740351916E-6}
          ($DA7B,$C158,$FD11,$8EFD,$BFEA),  {-5.3268724226180054113E-7}
          ($B0F6,$C479,$89AE,$93F2,$3FE5),  {+1.7223345741053954363E-8}
          ($005A,$13AE,$DF56,$E429,$3FE0),  {+8.3005421071190512062E-10}
          ($3840,$FF67,$B4A5,$E815,$BFDD),  {-1.0553994239956561153E-10}
          ($B1C8,$BD78,$A291,$8AAE,$3FD9),  {+3.9415842852183794790E-12}
          ($AD9E,$0DAE,$A912,$A30F,$3FD2),  {+3.6206853862057649821E-14}
          ($0FA5,$C7C9,$0F38,$C18C,$BFD0),  {-1.0744022741827186475E-14}
          ($8F70,$F721,$8BBB,$9020,$3FCC),  {+5.0004143941936846223E-16}
          ($38FD,$009F,$EE9C,$E667,$BFC5),  {-6.2451667275159614503E-18}
          ($88DA,$7509,$6FFB,$9909,$BFC2),  {-5.1850906707471146260E-19}
          ($4FF8,$4609,$2514,$A3C5,$3FBE)); {+3.4679666990622142954E-20}
var
  ga: array[0..gan-1] of extended absolute gah;

  function csfun(x: extended): extended;
  begin
    csfun := CSEvalX(2.0*x-1, ga, gan);
  end;

begin
  {See Temme [26] function auxgam for the ranges. Temme uses a rational}
  {approximation, I use a Chebyshev expansion calculated with Maple V. }
  if x<=-1 then igam_aux := -0.5
  else if x<0.0  then igam_aux := -(1.0+sqr(1.0+x)*csfun(x+1.0))/(1.0-x)
  else if x<=1.0 then igam_aux := csfun(x)
  else if x<=2.0 then igam_aux := ((x-2.0)*csfun(x)-1.0)/sqr(x)
  else igam_aux := (1.0/sfc_gamma(x+1.0)-1.0)/(x*(x-1.0));
end;


{---------------------------------------------------------------------------}
procedure igam_qtaylor(a,x: extended; var p,q: extended);
  {-Temme/Gautschi code for Q(a,x) when x < 1}
var
  r,s,t,v,u: extended;
begin
  {Ref: Temme [26] function qtaylor}
  q := powm1(x,a);
  s := -a*(a-1.0)*igam_aux(a);
  u := s - q*(1-s);
  p := a*x;
  q := a + 1.0;
  r := a + 3.0;
  t := 1.0;
  v := 1.0;
  repeat
    p := p + x;
    q := q + r;
    r := r + 2.0;
    t := -p*t/q;
    v := v + t;
  until abs(t) < eps_x*abs(v);
  t := a*(1.0-s)*x*exp(a*ln(x));
  r := t*v/(a+1);
  q := r + u;
  p := 1.0 - q;
end;


{---------------------------------------------------------------------------}
procedure igam_ptaylor(a,x,dax: extended; var p,q: extended);
  {-Temme/Gautschi code for P(a,x), dax = x^a*exp(-x)/gamma(a+1)}
var
  c,r: extended;
begin
  {Ref: Temme [26] formula (5.5) and function ptaylor}
  if (a<=0.0) or (x<=0.0) then p := 0.0
  else begin
    r := a;
    c := 1.0;
    p := 1.0;
    repeat
      r := r+1.0;
      c := c*x/r;
      p := p+c;
    until c < p*eps_x;
    p := p*dax/a;
  end;
  q := 1.0 - p;
end;


{---------------------------------------------------------------------------}
procedure igam_qfraction(a,x,dax: extended; var p,q: extended);
  {-Temme/Gautschi continued fraction for Q(a,x)}
var
  g,r,s,t,y,rho,tau: extended;
begin
  {Ref: Temme [26] function qfraction}
  s := 1.0-a;
  y := x+s;
  p := 0.0;
  q := y*(x-1.0-a);
  r := 4.0*y;
  t := 1.0;
  g := 1.0;
  rho:= 0.0;
  repeat
    p := p + s;
    q := q + r;
    r := r + 8.0;
    s := s + 2.0;
    tau := p*(1.0+rho);
    rho := tau/(q-tau);
    t := rho*t;
    g := g + t;
  until abs(t) < eps_x*abs(g);
  q := g/y*dax;
  p := 1.0-q;
end;


{---------------------------------------------------------------------------}
function igammaq_int(a, x: extended): extended;
  {-Return Q(a,x) when is an integer, a < min(30,x+1)}
var
  s,t: extended;
  n: integer;
begin
  {Ref: Boost [19], gamma.hpp, function finite_gamma_q}
  s := exp(-x);
  if (s>0.0) and (a>=2.0) then begin
    t := s;
    n := 1;
    while n<a do begin
      t := (t/n)*x;
      s := s+t;
      inc(n);
    end;
  end;
  igammaq_int := s;
end;


{---------------------------------------------------------------------------}
function igammaq_half(a, x: extended): extended;
  {-Return Q(a,x) when is a half-integer, a < min(30,x+1)}
var
  e,s,t: extended;
  n: integer;
begin
  {Ref: Boost [19], gamma.hpp, function finite_half_gamma_q}
  e := sfc_erfc(sqrt(x));
  if (e<>0.0) and (a>1.0) then begin
    t := 2.0*exp(-x)*sqrt(x/Pi);
    s := t;
    n := 2;
    while n<a do begin
      t := (t/(n-0.5))*x;
      s := s+t;
      inc(n);
    end;
    e := e + s;
  end;
  igammaq_half := e;
end;


{$ifdef VER50}
{---------------------------------------------------------------------------}
function sfc_igprefix(a,x: extended): extended;
  {-return dax = x^a*exp(-x)/gamma(a)}
var
  mu,d,t: extended;
begin
  mu := (x-a)/a;
  {Calculate dax = x^a*exp(-x)/gamma(a+1). This is the most sensible part for}
  {controlling the rounding errors occurring in the incomplete gamma functions}
  {This is function daxw cracked-down for TP5's brain-damaged FPU usage}
  if (a > 8.0) and (abs(mu) < 0.99) then begin
    t := ln1pmx(mu);
    d := a*t;
    t := 0.5*ln(a/TwoPi);
    d := d + t;
    t := sfc_lngcorr(a);
    d := d-t;
    d := exp(d);
  end
  else begin
    if (a > 8.0) and (x > 10.0) then begin
      t := mu - ln(x/a);
      d := a*t;
      t := sfc_lngcorr(a);
      d := d + t;
      t := 0.5*ln(a/TwoPi);
      d := d - t;
      d := exp(-d);
    end
    else if x > 10.0 then begin
      d := sfc_lngamma(a);
      t := ln(x);
      d := a*t-x-d;
      d := exp(d);
    end
    else begin
      {d := power(x,a)*exp(-x)/sfc_gamma(a);}
      d := power(x,a);
      t := exp(-x);
      d := d*t;
      t := sfc_gamma(a);
      d := d/t;
    end;
  end;
  sfc_igprefix := d;
end;

{$else}

{---------------------------------------------------------------------------}
function sfc_igprefix(a,x: extended): extended;
  {-return x^a*exp(-x)/gamma(a) using Lanczos sum}
const
  e1h: THexExtW = ($BE35,$6779,$B1B1,$BC5A,$3FFD);  {+3.6787944117144232158E-1}
var
  prefix,agh,alx,amx,amxa,d,mina,maxa: extended;
  em1: extended absolute e1h;  {1/e}
begin
  {Ref: Boost [19], gamma.hpp, function regularised_gamma_prefix}
  if a<1.0 then begin
    if x<Ln_MaxExt then sfc_igprefix := power(x,a)*exp(-x)/sfc_gamma(a)
    else begin
      {underflow exp(-x)}
      d := sfc_lngamma(a);
      prefix := a*ln(x) - x - d;
      sfc_igprefix := exp(prefix);
    end;
    exit;
  end;

  agh := a + lanczos_gm05;
  d := ((x-a) - lanczos_gm05)/agh;
  if ((abs(d*d*a) <= 100.0) and (a > 150.0)) then begin
    prefix := a*ln1pmx(d) - x*lanczos_gm05/agh;
    prefix := exp(prefix);
  end
  else begin
    alx := a * ln(x/agh);
    amx := a - x;
    if alx < amx then begin
      mina := alx;
      maxa := amx;
    end
    else begin
      maxa := alx;
      mina := amx;
    end;
    if (mina <= ln_MinExt) or (maxa >= ln_MaxExt) then begin
      amxa := amx/a;
      if (0.5*mina > ln_MinExt) and (0.5*maxa < ln_MaxExt) then begin
        {compute square root of the result and then square it}
        prefix := power(x/agh, 0.5*a)*exp(0.5*amx);
        prefix := sqr(prefix);
      end
      else if (0.25*mina > ln_MinExt) and (0.25*maxa < ln_MaxExt) and (x > a) then begin
        {compute the 4th root of the result then square it twice}
        prefix := power(x/agh, 0.25*a)*exp(0.25*amx);
        prefix := sqr(sqr(prefix));
      end
      else if ((amxa > ln_MinExt) and (amxa < ln_MaxExt)) then begin
        prefix := power((x*exp(amxa))/agh, a);
      end
      else prefix := exp(alx + amx);
    end
    else begin
      prefix := power(x/agh,a)*exp(amx);
    end;
  end;
  sfc_igprefix := prefix*sqrt(agh*em1)/lanczos(a, true);
end;
{$endif}


{---------------------------------------------------------------------------}
procedure sfc_incgamma_ex(a,x: extended; var p,q,dax: extended; calcdax: boolean);
  {-Return the normalised incomplete gamma functions P and Q, a>=0, x>=0}
  { P(a,x) = integral(exp(-t)*t^(a-1), t=0..x  )/gamma(a)}
  { Q(a,x) = integral(exp(-t)*t^(a-1), t=x..Inf)/gamma(a)}
  { If calcdax=true, dax = x^a*exp(-x)/gamma(a) is returned. This is extra}
  { effort only in the special cases, otherwise dax is calculated anyway.}
var
  mu,alfa: extended;
  use_temme: boolean;
begin
  dax := 0.0;
  if (x<0.0) or (a<0.0) or IsNanOrInf(x) or IsNanOrInf(a) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    p := NaN_x;
    q := NaN_x;
    exit;
  end;

  {The logic is based on Temme [26] and Boost's gamma.hpp [19]}
  {First handle trivial cases}
  if (a=0.0) and (x=0.0) then begin
    p := 0.5;
    q := 0.5;
    exit;
  end
  else if x=0.0 then begin
    p := 0.0;
    q := 1.0;
    exit;
  end
  else if a=0.0 then begin
    p := 1.0;
    q := 0.0;
    exit;
  end;

  {Calculate prefix factor if requested. The Lanczos based routine seems to}
  {give the best overall results, although daxw is sometimes more accurate.}
  if calcdax then dax := sfc_igprefix(a,x);

  if frac(2.0*a)=0.0 then begin
    {a is integer or half integer}
    if a=0.5 then begin
      {q = erfc(sqrt(x))}
      {p =  erf(sqrt(x))}
      x := sqrt(x);
      if x < 1.0 then begin
        p := sfc_erf(x);
        q := 1.0 - p;
      end
      else begin
        q := sfc_erfc(x);
        p := 1.0 - q;
      end;
      exit;
    end
    else if a=1.0 then begin
      q := exp(-x);
      p := -expm1(-x);
      exit;
    end
    else if (a < 30.0) and (a < x+1.0) then begin
      {Note: }
      if frac(a)=0.0 then begin
        {a is integer}
        if x > 0.6 then begin
          q := igammaq_int(a,x);
          p := 1.0 - q;
        end;
      end
      else begin
        {a is half-integer}
        if x > 0.2 then begin
          q := igammaq_half(a,x);
          p := 1.0 - q;
        end;
      end;
      exit;
    end;
  end;

  {Check if asymptotic expansion should be used}
  if a > 20.0 then begin
    mu := (x-a)/a;
    use_temme := false;
    if a > 200.0 then begin
      if abs(mu) < sqrt(20.0/a) then use_temme := true;
    end
    else if abs(mu) < 0.4 then use_temme := true;
    if use_temme then begin
      igam_pqasymp(a,x,p,q);
      exit;
    end;
  end;

  {If calcdax=true this has already been done above}
  if not calcdax then dax := sfc_igprefix(a,x);

  if dax < MinExtended then begin
    {Special case:  dax is zero or denormal. Note that this will not}
    {really help if the un-normalised functions shall be calculated!}
    if a > x then p := 0.0 else p := 1.0;
    q := 1.0-p;
  end
  else begin
    {Calculate Taylor series or continued fraction}
    if x > 0.25 then alfa := x+0.25 else alfa := -ln2/ln(x);
    if a > alfa then igam_ptaylor(a,x,dax,p,q)
    else if x < 1.0 then igam_qtaylor(a,x,p,q)
    else igam_qfraction(a,x,dax,p,q);
  end;
end;


{---------------------------------------------------------------------------}
procedure sfc_incgamma(a,x: extended; var p,q: extended);
  {-Return the normalised incomplete gamma functions P and Q, a>=0, x>=0}
  { P(a,x) = integral(exp(-t)*t^(a-1), t=0..x  )/gamma(a)}
  { Q(a,x) = integral(exp(-t)*t^(a-1), t=x..Inf)/gamma(a)}
var
  dax: extended;
begin
  sfc_incgamma_ex(a,x,p,q,dax,false);
end;


{---------------------------------------------------------------------------}
function sfc_igammap(a,x: extended): extended;
  {-Return the normalised lower incomplete gamma function P(a,x), a>=0, x>=0}
  { P(a,x) = integral(exp(-t)*t^(a-1), t=0..x)/gamma(a)}
var
  p,q,dax: extended;
begin
  sfc_incgamma_ex(a,x,p,q,dax,false);
  sfc_igammap := p;
end;


{---------------------------------------------------------------------------}
function sfc_igammaq(a,x: extended): extended;
  {-Return the normalised upper incomplete gamma function Q(a,x), a>=0, x>=0}
  { Q(a,x) = integral(exp(-t)*t^(a-1), t=x..Inf)/gamma(a)}
var
  p,q,dax: extended;
begin
  sfc_incgamma_ex(a,x,p,q,dax,false);
  sfc_igammaq := q;
end;


{---------------------------------------------------------------------------}
function sfc_igamma(a,x: extended): extended;
  {-Return the non-normalised upper incomplete gamma function}
  { GAMMA(a,x) = integral(exp(-t)*t^(a-1), t=x..Inf), x>=0}
var
  p,q,t,lnx: extended;
  k: integer;
  ok: boolean;
begin
  if (x<0.0) or IsNanOrInf(x) or IsNanOrInf(a) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_igamma := NaN_x;
    exit;
  end;
  {Special cases: GAMMA(a,0) = gamma(a) and GAMMA(0,x) = E_1(x)}
  if x=0.0 then sfc_igamma := sfc_gamma(a)
  else if (a=0.0) and (x>0.0) then sfc_igamma := sfc_e1(x)
  else if (a>0) and (a<MAXGAMX) then begin
    {Result is just Q(a,x)*Gamma(a)}
    sfc_incgamma_ex(a,x,p,q,t,false);
    t := sfc_gamma(a);
    sfc_igamma := q*t;
  end
  else if a<0.0 then begin
    {Logic for a<0 is similar to GSL [21], function gamma_inc.c}
    if (x>0.25) or (a <= -10000) then begin
      t := a*ln(x)-x;
      if t>ln_MaxExt then begin
        sfc_igamma := PosInf_x;
        exit;
      end;
      igam_qfraction(a,x,exp(t),p,q);
      sfc_igamma := q;
    end
    else if a >= -0.5 then begin
      igam_qtaylor(a,x,p,q);
      t := sfc_gamma(a);  {a <> 0!}
      sfc_igamma := q*t;
    end
    else begin
      {-10000 < a < -0.5}
      t := a - floor(a);
      {0 <= t < 1, set q = GAMMA(t,x) and use the recurrence formula}
      {NIST[30] 8.8.2:  GAMMA(t,x) = [GAMMA(t+1,x) - x^t*exp(-x)]/t }
      if t>0.0 then q := sfc_igammaq(t,x)*sfc_gamma(t)
      else q := sfc_e1(x);
      lnx := ln(x);
      for k:=1 to -floor(a) do begin
        t := t-1.0;
        p := -x + t*lnx;
        if p>ln_MaxExt then begin
          sfc_igamma := PosInf_x;
          exit;
        end;
        q := (q-exp(p))/t;
      end;
      sfc_igamma := q;
    end
  end
  else begin
    {here a >= MAXGAMX, i.e. gamma(a) overflows}
    if (x >= 1.0) and (a < x+0.25) then begin
      {qfraction if t=x^a*exp(-x) is not too large}
      t := a*ln(x);
      ok := false;
      if (t < ln_MaxExt) and (-x > ln_MinExt) then begin
        t := power(x,a)*exp(-x);
        ok := true;
      end
      else if t-x < ln_MaxExt then begin
        t := power(x/exp(x/a), a);
        ok := true;
      end;
      if ok then begin
        igam_qfraction(a,x,t,p,q);
        sfc_igamma := q;
        exit;
      end;
    end;
    {return Q(a,x)*Gamma(a) using lngamma}
    sfc_incgamma_ex(a,x,p,q,t,false);
    if q=0.0 then sfc_igamma := 0.0
    else begin
      t := sfc_lngamma(a);
      if q<0.75 then t := t + ln(q)
      else t := t + ln1p(-p);
      if t>ln_MaxExt then sfc_igamma := PosInf_x
      else sfc_igamma := exp(t);
    end;
  end;
end;


{---------------------------------------------------------------------------}
function igaml_mser(a,x: extended; var err: extended): extended;
  {-Return igaml_mser(a,x) = M(1,1+a,x) = igammal(a,x)*a*exp(x)/x^a}
var
  p,c,r,t,e: extended;
  n: integer;
begin
  {Modified non-normalised Temme/Gautschi code for P(a,x), cf. igam_ptaylor}
  r := a;
  c := 1.0;
  p := 1.0;
  e := 1.0;
  n := 0;
  repeat
    r := r+1.0;
    c := c*x/r;
    p := p+c;
    t := abs(c);
    if t>e then e := t;
    inc(n);
  until t < abs(p)*eps_x;
  e := e*eps_x;
  if p<>0 then e := e/abs(p);
  err := 0.5*n*eps_x + e;
  igaml_mser := p;
end;


{---------------------------------------------------------------------------}
function sfc_igammal(a,x: extended): extended;
  {-Return the non-normalised lower incomplete gamma function}
  { gamma(a,x) = integral(exp(-t)*t^(a-1), t=0..x); x>=0, a<>0,-1,-2,..}
var
  r,t,p: extended;
begin
  if (x<0.0) or IsNanOrInf(x) or IsNanOrInf(a) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_igammal := NaN_x;
    exit;
  end;
  if x=0.0 then sfc_igammal := 0.0
  else if a<=0.0 then begin
    if frac(a)=0.0 then begin
      {$ifopt R+}
        if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      {$endif}
      sfc_igammal := NaN_x;
      exit;
    end;
    {Split for TP5}
    r := sfc_gamma(a);
    t := sfc_igamma(a,x);
    p := r-t;
    {$ifdef debug}
      if abs(p)<sqrt_epsh*(abs(t)+abs(r)) then begin
        sfc_write_debug_str('*** sfc_igammal: loss of accuracy');
      end;
    {$endif}
    sfc_igammal := p;
  end
  else if (a<MinExtended) then sfc_igammal := PosInf_x
  else if (a<MinExtended*x) then sfc_igammal := 1.0/a
  else if (x<a-0.25) or (x<4.0) then begin
    if a<MAXGAMX then begin
      {Result is just P(a,x)*Gamma(a) if P(a,x) <> 0}
      p := sfc_igammap(a,x);
      if p<>0.0 then begin
        r := sfc_gamma(a);
        sfc_igammal := p*r;
        exit;
      end;
    end;
    p := igaml_mser(a,x,t);
    {$ifdef debug}
      if abs(p)<t then begin
        sfc_write_debug_str('*** sfc_igammal: large error from igaml_mser');
      end;
    {$endif}
    {Now gamma(a,x) = x^a*exp(-x)*p/a, but handle possible overflows}
    t := a*ln(x)-x;
    r := t + ln(p/a);
    if r > ln_MaxExt then sfc_igammal := PosInf_x
    else sfc_igammal := exp(r);
  end
  else begin
    {x>=4 and x >= a - 0.25}
    if a<MAXGAMX then begin
      {Split for TP5}
      r := sfc_gamma(a);
      t := sfc_igamma(a,x);
      p := r-t;
      {$ifdef debug}
        if abs(p)<sqrt_epsh*(abs(t)+abs(r)) then begin
          sfc_write_debug_str('*** sfc_igammal: loss of accuracy');
        end;
      {$endif}
      sfc_igammal := p;
    end
    else begin
      {Handle small window below overflow}
      sfc_igammal := PosInf_x;
      p := sfc_igammap(a,x);
      if p<>0 then begin
        t := ln(p) + sfc_lngamma(a);
        if t<=ln_MaxExt then sfc_igammal := exp(t);
      end;
    end;
  end;
end;


{---------------------------------------------------------------------------}
function lnigam(a,x,alx: extended): extended;
  {-Return ln(igamma(a,x)) for large x and for a <= x, SLATEC d9lgic, alx=ln(x)}
var
  eps,k,p,r,s,t,xma,xpa: extended;
const
  KMAX = 300;
begin
  xpa := x + 1.0 - a;
  xma := x - 1.0 - a;
  r   := 0.0;
  p   := 1.0;
  s   := p;
  k   := 0.0;
  eps := 0.25*eps_x;
  repeat
    k := k + 1.0;
    t := k*(a-k)*(1.0+r);
    r := -t/((xma+2.0*k)*(xpa+2.0*k)+t);
    p := r*p;
    s := s + p;
  until (abs(p) < eps*abs(s)) or (k>KMAX);
  lnigam := a*alx - x + ln(s/xpa);
  {$ifopt R+}
    if (k>KMAX) and (RTE_NoConvergence>0) then RunError(byte(RTE_NoConvergence));
  {$endif}
end;


{---------------------------------------------------------------------------}
function sfc_git(a,x: extended): extended;
  {-Return Tricomi's entire incomplete gamma function, x>=0,}
  { gammastar(a,x) = igammal(a,x)/gamma(a)/x^a = P(a,x)/x^a }
var
  p,r,t,h: extended;
  s: integer;
const
  alneps = 44.3614195558364998; {-ln(0.5*eps_x)}
begin
  if (x<0.0) or IsNanOrInf(x) or IsNanOrInf(a) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_git := NaN_x;
    exit;
  end;
  if x=0.0 then begin
    sfc_git := sfc_rgamma(a+1.0);
    exit;
  end
  else if a=0.0 then begin
    sfc_git := 1.0;
    exit;
  end;

  if a>=0 then begin
    if (a > x+1.0) and (a+1.0 < MAXGAMX) then begin
      p := igaml_mser(a,x,t);
      {$ifdef debug}
        if abs(p)<t then begin
          sfc_write_debug_str('*** sfc_git: large error from igaml_mser');
        end;
      {$endif}
      r := exp(-x);
      t := sfc_gamma(a+1.0);
      sfc_git := (p*r)/t;
    end
    else begin
      p := sfc_igammap(a,x);
      t := power(x,-a);
      sfc_git := t*p;
    end;
  end
  else if frac(a)=0 then begin
    {NIST[30], 8.4.12}
    sfc_git := power(x,-a);
  end
  else begin
    {Here a < 0}
    if x < 1.0 then begin
      p := igaml_mser(a,x,t);
      {$ifdef debug}
        if abs(p)<t then begin
          sfc_write_debug_str('*** sfc_git: large error from igaml_mser');
        end;
      {$endif}
      if a > -MAXGAMX then begin
        r := exp(-x);
        t := sfc_gamma(a+1.0);
        sfc_git := (p*r)/t;
      end
      else begin
        t := sfc_lngammas(a+1.0,s);
        if p>0.0 then p := ln(p)
        else begin
          s := -s;
          p := ln(abs(p));
        end;
        t := p-x-t;
        sfc_git := exp(t)*s;
      end;
    end
    else begin
      {Here a<0, x >= 1}
      if a < -MAXGAMX then begin
        {Compute result using logarithmic form, see SLATEC, dgamit.f label 40}
        r := ln(x);
        p := lnigam(a,x,r);
        t := sfc_lngammas(a+1.0,s);
        t := ln(abs(a)) + p - t;
        if t <= alneps then begin
          h := 1.0;
          if t >= -alneps then begin
            h := 1.0 + s*exp(t);
            if abs(h)<sqrt_epsh then begin
              {$ifdef debug}
                {ex: sfc_git(-1755.5,487.92898458757395315633918)}
                sfc_write_debug_str('*** sfc_git: loss of accuracy');
              {$endif}
            end;
          end;
          t := -a*r + ln(abs(h));
          if t>ln_Maxext then t := PosInf_x else t := exp(t);
          if h<0.0 then t := -t;
          sfc_git := t;
        end
        else begin
          t := t-a*r;
          if t>ln_Maxext then t := PosInf_x else t := exp(t);
          sfc_git := s*t;
        end;
      end
      else begin
        p := sfc_igamma(a,x);
        t := sfc_gamma(a);
        r := t-p;
        {$ifdef debug}
          if abs(r)<sqrt_epsh*abs(t) then begin
            sfc_write_debug_str('*** sfc_git: loss of accuracy');
          end;
        {$endif}
        p := r/t;
        t := power(x,-a);
        sfc_git := t*p;
      end;
    end;
  end;
end;


{---------------------------------------------------------------------------}
{----------------- Inverse Incomplete Gamma function -----------------------}
{---------------------------------------------------------------------------}

{---------------------------------------------------------------------------}
function DM_Fettis(a,lnB: extended): extended;
  {-Initial x for inverse incomplete gamma when a<1, q*Gamma(a) < 0.01}
var
  y,c1,c2,c3,c4,c5,a1,a2,a3: extended;
begin
  {DiDonato & Morris [27], Eq (25)}
  y  := -lnB;
  a1 := a-1.0;
  a2 := sqr(a);
  a3 := a2*a;
  c1 := a1*ln(y);
  c2 := 1.0 + c1;
  c3 := (-0.5*c1 + (a-2.0))*c1 + (1.5*a-2.5);
  c4 := ((c1/THREE - (1.5*a-2.5))*c1 + (a2 -6.0*a + 7.0))*c1 + (11.0*a2 - 46.0*a+47.0)/6.0;
  c5 := (((-0.25*c1 + (11.0*a - 17.0)/6.0)*c1 + (-3.0*a2 + 13.0*a - 13.0))*c1
           + (a3 - 12.5*a2 + 36.0*a - 30.5))*c1
           + (25.0*a3 - 195.0*a2 + 477.0*a - 379.0)/12.0;
  DM_Fettis :=((((c5/y + c4)/y + c3)/y + c2)*a1/y + c1) + y;
end;


{---------------------------------------------------------------------------}
function DM_CF6(a,p,q: extended): extended;
  {-Cornish-Fisher 6-term approximation for inverse incomplete gamma}
const
  ca: array[0..3] of extended = (3.31125922108741, 11.6616720288968, 4.28342155967104, 0.213623493715853);
  cb: array[0..4] of extended = (1.0, 6.61053765625462, 6.40691597760039, 1.27364489782223, 0.3611708101884203e-1);
var
  rta,s,s2,t: extended;
begin
  {DiDonato & Morris [27], Eq (32)}
  if p<0.5 then t := sqrt(-2.0*ln(p))
  else t := sqrt(-2.0*ln(q));
  s2 := PolEvalX(t,ca,4);
  s  := PolEvalX(t,cb,5);
  s  := t - s2/s;
  if p<0.5 then s := -s;
  {DiDonato & Morris [27], Eq (31)}
  s2  := s*s;
  rta := sqrt(a);
  t   := a + s*rta + (s2 - 1.0)/3.0 + s*(s2 - 7.0)/(36.0*rta)
           - ((3.0*s2 + 7.0)*s2 - 16.0)/(810.0*a)
           + s*((9.0*s2 + 256.0)*s2 - 433.0)/(38880.0*a*rta);
  if t<0.0 then DM_CF6 := 0.0
  else DM_CF6 := t;
end;


{---------------------------------------------------------------------------}
function igamma_inv_guess(a,p,q: extended; var noiter: boolean): extended;
  {-Return an initial approximation for inverse incomplete gamma}
var
  b,g,s,t,u,v,w,x,y,z,am1,ap1: extended;
  i: integer;
const
  ln10 = 2.302585092994;
begin
  noiter := false;
  am1 := a-1.0;
  if a=1.0 then x := - ln(q)
  else if a<1.0 then begin
    g := sfc_gamma(a);
    b := q*g;
    if ((b > 0.6) or ((b >= 0.45) and (a >= 0.3))) then begin
      {DiDonato & Morris [27], Eq (21), see comment in Boost for a slight variation}
      if (b*q > 1e-8) and (q > 1e-5) then u := power(p*g*a, 1.0/a)
      else u := exp(-q/a - EulerGamma);
      x := u/(1.0 - u/(a+1.0));
    end
    else if (a < 0.3) and (b >= 0.35) then begin
      {DiDonato & Morris [27], Eq (22)}
      t := exp(-EulerGamma - b);
      u := t*exp(t);
      x := t*exp(u);
    end
    else if (b > 0.15) or (a >= 0.3) then begin
      {DiDonato & Morris [27], Eq (23)}
      y := -ln(b);
      v := y + am1*ln(y);
      x := y + am1*ln(v) - ln1p(-am1/(1.0+v));
    end
    else if b>0.1 then begin
      {DiDonato & Morris [27], Eq (24)}
      y := -ln(b);
      v := y + am1*ln(y);
      t := v*v + 2.0*(3.0-a)*v + (2.0-a)*(3.0-a);
      u := v*v + (5.0-a)*v + 2.0;
      x := y + am1*ln(v) - ln(u/t);
    end
    else begin
      {DiDonato & Morris [27], Eq (25)}
      x  := DM_Fettis(a,ln(b));
    end;
  end
  else begin {a > 1}
    w := DM_CF6(a,p,q);
    t := abs(1.0 - w/a);
    if (a >= 500.0) and (t < 1e-6) then begin
      x := w;
      if t<1e-17 then noiter := true;
    end
    else if p>0.5 then begin
      if w < 3.0*a then begin
        x := w;
      end
      else begin
        {here b = ln(q*Gamma(a)) = ln(B)}
        b := ln(q) + sfc_lngamma(a);
        t := maxx(2.0, a*am1);
        if b < -ln10*t then begin
          x := DM_Fettis(a,b)
        end
        else begin
          {DiDonato & Morris [27], Eq (33)}
          u := am1*ln(w) - b - ln1p(-am1/(1.0+w));
          x := am1*ln(u) - b - ln1p(-am1/(1.0+u));
        end;
      end
    end
    else begin
      {p < 0.5}
      z := w;
      ap1 := a+1.0;
      if w < 0.15*ap1 then begin
        {DiDonato & Morris [27], Eq (35)}
        v := ln(p) + sfc_lngamma(ap1);
        t := a+2.0;
        z := exp((v+w)/a);
        s := ln1p(z/ap1*(1.0 + z/t));
        z := exp((v + z - s)/a);
        {Note: the next line is missing in the Boost code!}
        s := ln1p(z/ap1*(1.0 + z/t));
        z := exp((v + z - s)/a);
        s := ln1p(z/ap1 * (1.0 + z/t*(1.0 + z/(a+3.0))));
        z := exp((v + z - s)/a);
      end;

      {WE: The setting noiter := true if z<0.002(a+1) has been dropped}
      if (z <= 0.01*ap1) or (z > 0.7*ap1) then begin
        x := z;
      end
      else begin
        {DiDonato & Morris [27], Eq (34)}
        i := 1;
        t := z/ap1;
        s := 1.0+t;
        while (i<100) and (t>1e-4) do begin
          inc(i);
          t := t*(z/(a+i));
          s := s+t;
        end;
        {DiDonato & Morris [27], Eq (36)}
        t := ln(p) + sfc_lngamma(ap1) - ln(s);
        z := exp((z+t)/a);
        x := z*(1.0 - (a*ln(z) - z - t)/(a-z));
      end;
    end;
  end;
  igamma_inv_guess := x;
end;


{---------------------------------------------------------------------------}
procedure sfc_incgamma_inv(a,p,q: extended; var x: extended; var ierr: integer);
  {-Return the inverse normalised incomplete gamma function, i.e. calculate}
  { x with P(a,x)=p and Q(a,x)=q. Input parameter a>0, p>=0, q>=0 and p+q=1.}
  { ierr is >= 0 for success, < 0 for input errors or iterations failures. }
var
  pn,qn,r,t,h,w,d,td,am1: extended;
  done: boolean;
const
  eps  = 1.1e-13;      {~ (eps_x/2)^(2/3), cf. Boost's [19] 2/3*digits}
  tol  = 1.1e-13;
  rmin = 3.101e-4913;  {Threshold for derivative, ~MinExtended/eps_x}
begin

  {The classical reference is A.R. Didonato, A.H. Morris[27]. The original}
  {text and equations are used together with the Boost interpretation from}
  {\boost\math\special_functions\detail\igamma_inverse.hpp [19], which is }
  {Copyright John Maddock 2006, see 3rdparty.ama for Boost license}

  {Some of the ierr return codes from [27] are used: }
  { ierr >=  0, iteration count                      }
  { ierr  = -2, if a <= 0                            }
  { ierr  = -4, if p < 0, q < 0, or |p+q-1| > eps_d  }
  { ierr  = -6, if 10 iterations were performed      }
  { ierr  = -7, iteration failed                     }
  { ierr  = -8, x is calculated with unknown accuracy}

  x := -1.0;
  if a <= 0.0 then begin
    ierr := -2;
    exit;
  end;

  {Use eps_d for check because function may be called with 'double' p,q}
  if (p<0.0) or (q<0.0) or (abs(p+q-1.0)>eps_d) then begin
    ierr := -4;
    exit;
  end;

  if (p=1.0) and (q=0.0) then begin
    x := PosInf_x;
    exit;
  end;

  x := 0.0;
  ierr := 0;
  if p=0.0 then exit;

  if a=1.0 then begin
    {No guess, no iteration. Use known answer}
    if q <= 0.9 then x := -ln(q)
    else x := -ln1p(-p);
    exit;
  end;

  {Calculate initial approximation, done if no iteration needed/useful}
  x := igamma_inv_guess(a,p,q,done);
  if done then begin
    exit;
  end;

  if (p <= MinExtended) or (q <= MinExtended) then begin
    ierr := -8;
    exit;
  end;

  td   := tol*minx(p,q);  {iteration tolerance for |p-pn| or |q-qn|}
  am1  := a - 1.0;
  done := false;

  repeat
    {Note that DiDonato & Morris [27] use two different iteration loops:}
    {one for P and one for Q, they can be combined with minimal effort. }
    inc(ierr);
    if ierr>10 then begin
      ierr := -6;
      exit;
    end;
    sfc_incgamma_ex(a,x,pn,qn,r,true);
    if r<rmin then begin
      ierr := -8;
      exit;
    end;
    if p<=0.5 then d := pn-p else d := q-qn;
    t := d/r;
    w := 0.5*(am1-x);
    if (abs(t) < 0.1) and (abs(w*t) <= 0.1) then begin
      {Schrder step}
      h := w*sqr(t);
      done := (abs(w) > 1.0) and (abs(h) < eps);
      h := t + h;
    end
    else begin
      {Newton step}
      h := t;
    end;
    x := x*(1.0-h);
    if x<=0.0 then begin
      ierr := -7;
      exit;
    end;
    if (abs(h) < eps) or (abs(d) < td) then done := true;
  until done;
end;


{---------------------------------------------------------------------------}
function sfc_igamma_inv(a,p,q: extended): extended;
  {-Return the inverse normalised incomplete gamma function, i.e. calculate}
  { x with P(a,x)=p and Q(a,x)=q. Input parameter a>0, p>=0, q>0 and p+q=1.}
var
  ierr: integer;
  x: extended;
begin
  sfc_incgamma_inv(a,p,q,x,ierr);
  sfc_igamma_inv := x;
  if ierr<0 then begin
    if (ierr=-2) or (ierr=-4) then begin
      {$ifopt R+}
        if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      {$endif}
      sfc_igamma_inv := NaN_x;
      exit;
    end;
    {$ifopt R+}
      if RTE_NoConvergence>0 then RunError(byte(RTE_NoConvergence));
    {$endif}
  end;
end;


{---------------------------------------------------------------------------}
function sfc_igammap_inv(a,p: extended): extended;
  {-Inverse incomplete gamma: return x with P(a,x)=p, a>=0, 0<=p<1}
begin
  sfc_igammap_inv := sfc_igamma_inv(a,p,1.0-p);
end;


{---------------------------------------------------------------------------}
function sfc_igammaq_inv(a,q: extended): extended;
  {-Inverse complemented incomplete gamma: return x with Q(a,x)=q, a>=0, 0<q<=1}
begin
  sfc_igammaq_inv := sfc_igamma_inv(a,1.0-q,q);
end;


{---------------------------------------------------------------------------}
function sfc_taylor(x: extended; n: integer): extended;
  {-Return the Taylor coefficient x^n/n!, n>=0}
var
  z,f: extended;
  k: integer;
const
  zmin = -10991.3977; { > ln(MinExtended*100!)}
begin
  if IsNanOrInf(x) or (n<0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_taylor := NaN_x;
    exit;
  end;
  if n<=0 then sfc_taylor := 1.0
  else if (n=1) or (x=0.0) then sfc_taylor := x
  else begin
    z := n*ln(abs(x));
    if z<ln_MinExt then sfc_taylor := 0.0
    else if (n<=100) and (z<ln_MaxExt) and (z>zmin) then begin
      {do not waste time for sfc_lnfac(n)}
      f := 1.0;
      for k:=1 to n do f := f*(x/k);
      sfc_taylor := f;
    end
    else begin
      f := sfc_lnfac(n);
      f := z-f;
      if f<ln_MinExt then sfc_taylor := 0.0
      else if f>ln_MaxExt then sfc_taylor := copysign(PosInf_x,x)
      else begin
        {no over/underflow}
        if n>2000 then begin
          f := exp(f);
          if (x<0.0) and odd(n) then f := -f;
        end
        else begin
          f := 1.0;
          for k:=1 to n do f := f*(x/k);
        end;
        sfc_taylor := f;
      end;
    end;
  end;
end;


end.
